from hashlib import sha256
from json import dumps as json_dumps
from typing import TYPE_CHECKING, Any, Type

from fastapi import Response
from pydantic import BaseModel, model_validator
from pydantic._internal._model_construction import ModelMetaclass
from pydantic.fields import Field, FieldInfo
from typing_extensions import dataclass_transform


class FieldClassDefinition(BaseModel):
    root_model: Type[BaseModel]
    key: str
    field_definition: FieldInfo

    model_config = {
        "arbitrary_types_allowed": True,
    }


@dataclass_transform(kw_only_default=True, field_specifiers=(Field,))
class ReturnModelMetaclass(ModelMetaclass):
    INTERNAL_RENDER_FIELDS = ["metadata"]

    if not TYPE_CHECKING:  # pragma: no branch
        # Following the lead of the pydantic superclass, we wrap with a non-TYPE_CHECKING
        # block: "otherwise mypy allows arbitrary attribute access""

        def __new__(
            self,
            cls_name: str,
            bases: tuple[type[Any], ...],
            namespace: dict[str, Any],
            *args,
            **kwargs: Any,
        ):
            # Pydantic uses exceptions in the __getattr__ to handle collection of fields
            # in set_model_fields. While we're still initializing the class we have no
            # need for our custom accessor logic - so we temporarily turn it off.
            self.is_constructing = True
            obj = super().__new__(self, cls_name, bases, namespace, *args, **kwargs)
            self.is_constructing = False
            return obj

        def __getattr__(self, key: str) -> Any:
            if self.is_constructing:
                return super().__getattr__(key)

            try:
                return super().__getattr__(key)
            except AttributeError:
                # Determine if this field is defined within the spec
                # If so, return it
                if key in self.model_fields and key not in self.INTERNAL_RENDER_FIELDS:
                    return FieldClassDefinition(
                        root_model=self,
                        key=key,
                        field_definition=self.model_fields[key],
                    )
                raise


class HashableAttribute(BaseModel):
    """
    Even with frozen=True, we can't hash our attributes because they include dictionary field types.
    Instead provide a mixin to calculate the hash of the current state of the attributes.

    """

    def __hash__(self):
        model_json = json_dumps(self.model_dump(), sort_keys=True)
        hash_object = sha256(model_json.encode())
        # __hash__ must return an integer
        return int(hash_object.hexdigest(), 16)


class MetaAttribute(HashableAttribute, BaseModel):
    name: str | None = None
    content: str | None = None
    optional_attributes: dict[str, str] = {}


class ThemeColorMeta(MetaAttribute):
    """
    Customizes the default color that is attached to the page.

    ```python
    ThemeColorMeta(
        color="white",
        media="light",
    )
    ```

    """

    color: str
    media: str | None = None

    @model_validator(mode="after")
    def create_attribute(self):
        self.name = "theme-color"
        self.content = self.color
        if self.media:
            self.optional_attributes = {"media": self.media}
        return self


class ViewportMeta(MetaAttribute):
    """
    Defines the bounds on the current page and how much users are able to zoom.

    ```python
    ViewportMeta(
        initial_scale=1.0,
        maximum_scale=2.0,
        user_scalable=True,
    )
    ```

    """

    width: str = "device-width"
    initial_scale: float = 1.0
    maximum_scale: float = 1.0
    user_scalable: bool = False

    @model_validator(mode="after")
    def create_attribute(self):
        user_scalable_str = "yes" if self.user_scalable else "no"

        self.name = "viewport"
        self.content = f"width={self.width}, initial-scale={self.initial_scale}, maximum-scale={self.maximum_scale}, user-scalable={user_scalable_str}"
        return self


class LinkAttribute(HashableAttribute, BaseModel):
    rel: str
    href: str
    optional_attributes: dict[str, str] = {}


class ScriptAttribute(HashableAttribute, BaseModel):
    src: str
    asynchronous: bool = False
    defer: bool = False
    optional_attributes: dict[str, str] = {}


class Metadata(BaseModel):
    """
    Metadata lets the client specify the different metadata definitions that should
    appear on the current page. These are outside the scope of React management so are
    only handled once on the initial page render.

    """

    title: str | None = None

    # Specify dynamic injection of tags into the <head>
    metas: list[MetaAttribute] = []
    links: list[LinkAttribute] = []
    scripts: list[ScriptAttribute] = []

    # Allows the client to specify a different response type
    # that should occur on initial render
    # Useful for redirects, adding cookies, etc.
    explicit_response: Response | None = None

    # If enabled, we won't attempt to use the global metadata for this route
    # Helpful for plugins or otherwise for nested routes that should escape the container
    ignore_global_metadata: bool = False

    model_config = {
        "extra": "forbid",
        "arbitrary_types_allowed": True,
    }


class RenderBase(BaseModel, metaclass=ReturnModelMetaclass):
    """
    Base class for all renderable data models. Subclass this model when defining
    your own component data schema.
    """

    metadata: Metadata | None = Field(default=None, exclude=True)

    model_config = {
        # Frozen parameters are required so we can hash the render values to check
        # for changes in our SSR renderer
        "frozen": True,
    }


class RenderNull(RenderBase):
    pass



from enum import Enum, IntEnum, StrEnum
from json import dumps as json_dumps
from typing import Generic, TypeVar

import pytest
from pydantic import BaseModel, Field, create_model

from mountaineer.client_builder.build_schemas import (
    OpenAPISchema,
    OpenAPIToTypescriptSchemaConverter,
)
from mountaineer.client_builder.openapi import OpenAPIProperty, OpenAPISchemaType

T = TypeVar("T")


class SubModel1(BaseModel):
    sub_a: str


class SubModel2(BaseModel):
    sub_b: int


class MyStrEnum(StrEnum):
    VALUE_1 = "value_1"
    VALUE_2 = "value_2"


class MyIntEnum(IntEnum):
    VALUE_1 = 1
    VALUE_2 = 2


class MyEnum(Enum):
    VALUE_1 = "value_1"
    VALUE_2 = 5


class MyModel(BaseModel):
    a: str = Field(description="The a field")
    b: int
    c: SubModel1
    d: list[SubModel1]
    both_sub: list[SubModel1 | SubModel2 | None]
    sub_map: dict[str, SubModel1 | None]


def test_basic_interface():
    converter = OpenAPIToTypescriptSchemaConverter()

    json_schema = OpenAPISchema(**converter.get_model_json_schema(MyModel))
    result = converter.convert_schema_to_typescript(json_schema)

    assert set(result.keys()) == {"MyModel", "SubModel1", "SubModel2", "Sub Map"}
    assert "interface MyModel {" in result["MyModel"]


def test_model_gathering_pydantic_models():
    """
    Ensure we are able to traverse a single model definition for all the
    sub-models it uses.

    """
    schema = OpenAPISchema(**MyModel.model_json_schema())

    converter = OpenAPIToTypescriptSchemaConverter()
    all_models = converter.gather_all_models(schema)

    # OpenAPI makes an object for the dictionary as well
    assert len(all_models) == 4
    assert {m.title for m in all_models} == {
        "SubModel1",
        "SubModel2",
        "MyModel",
        "Sub Map",
    }


def test_model_gathering_enum_models():
    class EnumModel(BaseModel):
        a: MyStrEnum
        b: MyIntEnum
        c: MyEnum

    schema = OpenAPISchema(**EnumModel.model_json_schema())

    converter = OpenAPIToTypescriptSchemaConverter()
    all_models = converter.gather_all_models(schema)

    assert len(all_models) == 4
    assert {m.title for m in all_models} == {
        "EnumModel",
        "MyStrEnum",
        "MyIntEnum",
        "MyEnum",
    }


@pytest.mark.parametrize(
    "python_type,expected_typescript_types",
    [
        (str, ["value: string"]),
        (int, ["value: number"]),
        (SubModel1, ["value: SubModel1"]),
        (list[SubModel1], ["value: Array<SubModel1>"]),
        (dict[str, SubModel1], ["value: Record<string, SubModel1>"]),
        (dict[str, int], ["value: Record<string, number>"]),
        (dict[str, dict[str, str]], ["value: Record<string, Record<string, string>>"]),
        ("SubModel1", ["value: SubModel1"]),
        (MyStrEnum, ["value: MyStrEnum"]),
        (MyIntEnum, ["value: MyIntEnum"]),
        (MyEnum, ["value: MyEnum"]),
    ],
)
def test_python_to_typescript_types(
    python_type: type, expected_typescript_types: list[str]
):
    """
    Test type resolution when attached to a given model's field
    """
    # Create a model schema automatically with the passed in typing
    fake_model = create_model(
        "FakeModel",
        value=(python_type, Field()),
    )

    schema = OpenAPISchema(**fake_model.model_json_schema())

    converter = OpenAPIToTypescriptSchemaConverter()
    interface_definition = converter.convert_schema_to_interface(
        schema,
        base=schema,
        defaults_are_required=False,
        all_fields_required=False,
    )

    for expected_str in expected_typescript_types:
        assert expected_str in interface_definition


def test_require_json_dictionaries():
    """
    JSON dictionaries can only serialize with string keys, so we need to make sure
    that the TypeScript interface enforces this.
    """

    class InvalidDictModel(BaseModel):
        value: dict[int, str]

    class ValidDictModel(BaseModel):
        value: dict[str, str]

    converter = OpenAPIToTypescriptSchemaConverter()
    converter.validate_typescript_candidate(ValidDictModel)

    with pytest.raises(ValueError):
        converter.validate_typescript_candidate(InvalidDictModel)


def test_get_model_json_schema_excludes_masked_fields():
    """
    Ensure we avoid serializing "excluded" fields.
    """

    class ExcludedModel(BaseModel):
        a: str

    class IncludedModel(BaseModel):
        b: str

    class MainModel(BaseModel):
        excluded_obj: ExcludedModel | None = Field(default=None, exclude=True)
        included_obj: IncludedModel | None

    builder = OpenAPIToTypescriptSchemaConverter()
    raw_openapi_result = json_dumps(MainModel.model_json_schema())
    fixed_openapi_result = json_dumps(builder.get_model_json_schema(MainModel))

    # First we check that pydantic actually does serialize excluded fields
    # If this starts failing, we should be able to remove our workaround
    assert "excluded_obj" in raw_openapi_result
    assert "ExcludedModel" in raw_openapi_result

    assert "excluded_obj" not in fixed_openapi_result
    assert "ExcludedModel" not in fixed_openapi_result
    assert "included_obj" in fixed_openapi_result
    assert "IncludedModel" in fixed_openapi_result


def test_format_enums():
    class MyModel(BaseModel):
        # String type enums
        a: MyStrEnum
        # Int type enums
        b: MyIntEnum
        # Mixed type enums: string and int
        c: MyEnum

    converter = OpenAPIToTypescriptSchemaConverter()
    json_schema = OpenAPISchema(**converter.get_model_json_schema(MyModel))
    js_interfaces = converter.convert_schema_to_typescript(json_schema)

    assert (
        js_interfaces["MyStrEnum"]
        == "enum MyStrEnum {\nValue1 = 'value_1',\nValue2 = 'value_2'\n}"
    )
    assert (
        js_interfaces["MyIntEnum"] == "enum MyIntEnum {\nValue__1 = 1,\nValue__2 = 2\n}"
    )
    assert (
        js_interfaces["MyEnum"] == "enum MyEnum {\nValue1 = 'value_1',\nValue__5 = 5\n}"
    )


def test_format_generics():
    class MyModel(BaseModel, Generic[T]):
        a: T

    converter = OpenAPIToTypescriptSchemaConverter()
    json_schema = OpenAPISchema(**converter.get_model_json_schema(MyModel[str]))
    js_interfaces = converter.convert_schema_to_typescript(json_schema)

    assert js_interfaces == {"MyModel[str]": "interface MyModelStr {\n  a: string;\n}"}


@pytest.mark.parametrize("defaults_are_required", [True, False])
def test_defaults_are_required(defaults_are_required: bool):
    class MyModelExplicitField(BaseModel):
        a: str = Field(default="default value")

    class MyModelImplicitField(BaseModel):
        a: str = "default value"

    # Both behaviors should be the same
    for base_model in [MyModelExplicitField, MyModelImplicitField]:
        converter = OpenAPIToTypescriptSchemaConverter()

        json_schema = OpenAPISchema(**converter.get_model_json_schema(base_model))
        js_interfaces = converter.convert_schema_to_typescript(
            json_schema, defaults_are_required=defaults_are_required
        )
        model_name = base_model.__name__

        if defaults_are_required:
            assert "a: string" in js_interfaces[model_name]
            assert "a?: string" not in js_interfaces[model_name]
        else:
            assert "a: string" not in js_interfaces[model_name]
            assert "a?: string" in js_interfaces[model_name]


@pytest.mark.parametrize(
    "model_title, expected_interface",
    [
        ("MyModel", "MyModel"),
        # We've seen cases where sub-variables are converted to multiple words
        ("My Model", "MyModel"),
        ("My model", "MyModel"),
    ],
)
def test_get_typescript_interface_name(model_title: str, expected_interface: str):
    converter = OpenAPIToTypescriptSchemaConverter()
    assert (
        converter.get_typescript_interface_name(
            OpenAPIProperty.from_meta(
                title=model_title,
                variable_type=OpenAPISchemaType.OBJECT,
            )
        )
        == expected_interface
    )



from dataclasses import dataclass

from pydantic import BaseModel

from mountaineer.annotation_helpers import yield_all_subtypes


def test_yield_all_subtypes():
    class SubModel(BaseModel):
        sub_a: int

    class SubModelOnlyReferenced(BaseModel):
        sub_a: int

    @dataclass
    class SubDataclass1:
        sub_a: int
        sub_dataclass_forwardref: "SubDataclass2"

    @dataclass
    class SubDataclass2:
        sub_a: int

    class ComplexModel(BaseModel):
        a: str
        b: SubModel
        c: list[SubModel]
        d_forwardref: "SubModelOnlyReferenced"
        e_forwardref: list["SubModelOnlyReferenced"]
        f_dataclass: SubDataclass1

    # We need to pass along the global namespace so we're able to resolve the forward references
    # in the dataclasses. Pydantic does this for us at runtime.
    assert set(yield_all_subtypes(ComplexModel, _locals=locals())) == {
        # Regular models
        ComplexModel,
        SubModel,
        SubDataclass1,
        SubDataclass2,
        SubModelOnlyReferenced,
        # Regular types
        str,
        int,
        # Origins + Arguments
        list[SubModel],
        list[SubModelOnlyReferenced],
        # Origins
        list,
    }



{% if create_stub_files %}
from uuid import UUID

from mountaineer import Metadata, RenderBase, ControllerBase, APIException, sideeffect
from mountaineer.database import DatabaseDependencies

from fastapi import Request, Depends
from sqlalchemy.ext.asyncio import AsyncSession
from fastapi.exceptions import HTTPException
from pydantic import BaseModel

from {{project_name}} import models

class NotFoundException(APIException):
    status_code = 404
    detail = "Detail item not found"


class UpdateTextRequest(BaseModel):
    description: str


class DetailRender(RenderBase):
    id: int
    description: str


class DetailController(ControllerBase):
    url = "/detail/{detail_id}/"
    view_path = "/app/detail/page.tsx"

    def __init__(self):
        super().__init__()

    async def render(
        self,
        detail_id: int,
        session: AsyncSession = Depends(DatabaseDependencies.get_db_session)
    ) -> DetailRender:
        detail_item = await session.get(models.DetailItem, detail_id)
        if not detail_item:
            raise NotFoundException()

        return DetailRender(
            id=detail_item.id,
            description=detail_item.description,
            metadata=Metadata(title=f"Detail: {detail_id}"),
        )

    @sideeffect
    async def update_text(
        self,
        detail_id: int,
        payload: UpdateTextRequest,
        session: AsyncSession = Depends(DatabaseDependencies.get_db_session)
    ) -> None:
        detail_item = await session.get(models.DetailItem, detail_id)
        if not detail_item:
            raise NotFoundException()

        detail_item.description = payload.description
        await session.commit()
{% endif %}


from abc import ABC, abstractmethod
from importlib.metadata import PackageNotFoundError
from inspect import getmembers, isawaitable, ismethod
from pathlib import Path
from re import compile as re_compile
from time import monotonic_ns
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    Iterable,
    Mapping,
    Optional,
    ParamSpec,
    cast,
)

from fastapi.responses import HTMLResponse
from inflection import underscore

from mountaineer.actions import (
    FunctionActionType,
    FunctionMetadata,
    get_function_metadata,
)
from mountaineer.config import get_config
from mountaineer.js_compiler.source_maps import SourceMapParser
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, resolve_package_path
from mountaineer.render import (
    LinkAttribute,
    MetaAttribute,
    Metadata,
    RenderBase,
    RenderNull,
    ScriptAttribute,
)
from mountaineer.ssr import V8RuntimeError, render_ssr

if TYPE_CHECKING:
    from mountaineer.app import ControllerDefinition

RenderInput = ParamSpec("RenderInput")


class ControllerBase(ABC, Generic[RenderInput]):
    url: str
    # Typically, view paths should be a relative path to the local
    # Paths are only used if you need to specify an absolute path to another
    # file on disk
    view_path: str | ManagedViewPath

    bundled_scripts: list[str]

    # Upon registration, the AppController will mount a wrapper
    # with state metadata
    definition: Optional["ControllerDefinition"] = None

    def __init__(
        self, slow_ssr_threshold: float = 0.1, hard_ssr_timeout: float | None = 10.0
    ):
        """
        One Controller should be created for every frontend page in your webapp. Clients can override
        this `__init__` function so long as they call `super().__init__()` at the start of their init
        to setup the internal handlers.

        :param slow_ssr_threshold: Each python process has a single V8 runtime associated with
        it, so SSR rendering can become a bottleneck if it requires processing. We log a warning
        if we detect that an SSR render took longer than this threshold.
        :param hard_ssr_timeout: If the SSR render takes longer than this threshold, we will
        automatically kill the V8 runtime and return an error to the client. This is useful for
        avoiding blocking the reset of the server process if the React render logic hangs.

        """
        # Injected by the build framework
        self.bundled_scripts: list[str] = []
        self.initialized = True
        self.slow_ssr_threshold = slow_ssr_threshold
        self.hard_ssr_timeout = hard_ssr_timeout
        self.source_map: SourceMapParser | None = None

        self.view_base_path: Path | None = None
        self.ssr_path: Path | None = None

        self.resolve_paths()

    @abstractmethod
    def render(
        self, *args: RenderInput.args, **kwargs: RenderInput.kwargs
    ) -> RenderBase | None | Coroutine[Any, Any, RenderBase | None]:
        """
        Render provides the raw data payload that will be sent to the frontend on initial
        render and during any sideeffect update. In most cases, you should return a RenderBase
        instance. If you have no data to display you can also return None.

        This function must be explicitly typehinted with your response type, which allows the
        AppController to generate the correct TypeScript types for the frontend:

        ```python
        class MyServerData(RenderBase):
            pass

        class MyController:
            def render(self) -> MyServerData:
                pass
        ```

        If you don't intend to sync any data from server->client you can typehint this function
        with an explicit None return annotation:

        ```python
        class MyController:
            def render(self) -> None:
                pass
        ```

        Render functions accept any number of arguments and keyword arguments, following the FastAPI
        route parameter style. This includes query parameters, path parameters, and request bodies.

        ```python
        class MyController:
            url = "/my-url/{path_param}"

            def render(
                self,
                query_param: str,
                path_param: int,
                dependency: MyDependency = Depends(MyDependency),
            ) -> MyServerData:
                ...
        ```

        :return: A RenderBase instance or None
        """
        pass

    async def _generate_html(self, *args, global_metadata: Metadata | None, **kwargs):
        # Because JSON is a subset of JavaScript, we can just dump the model as JSON and
        # insert it into the page.
        server_data = self.render(*args, **kwargs)
        if isawaitable(server_data):
            server_data = await server_data
        if server_data is None:
            server_data = RenderNull()

        # This isn't expected to happen, but we add a check to typeguard the following logic
        if not isinstance(server_data, RenderBase):
            raise ValueError(
                f"Controller.render() must return a RenderBase instance, not {type(server_data)}"
            )

        # If we got back metadata that includes a redirect, we should short-circuit the rest of the
        # render process and return a redirect response
        if server_data.metadata and server_data.metadata.explicit_response:
            return server_data.metadata.explicit_response

        metadatas: list[Metadata] = []
        if server_data.metadata:
            metadatas.append(server_data.metadata)
        if global_metadata and (
            server_data.metadata is None
            or not server_data.metadata.ignore_global_metadata
        ):
            metadatas.append(global_metadata)

        header_str = "\n".join(self._build_header(self._merge_metadatas(metadatas)))

        ssr_html = self._generate_ssr_html(server_data)

        # Client-side react scripts that will hydrate the server side contents on load
        server_data_json = server_data.model_dump_json()
        optional_scripts = "\n".join(
            [
                f"<script src='/static/{script_name}'></script>"
                for script_name in self.bundled_scripts
            ]
        )

        page_contents = f"""
        <html>
        <head>
        {header_str}
        </head>
        <body>
        <div id="root">{ssr_html}</div>
        <script type="text/javascript">
        const SERVER_DATA = {server_data_json};
        </script>
        {optional_scripts}
        </body>
        </html>
        """

        return HTMLResponse(page_contents)

    def _generate_ssr_html(self, server_data: RenderBase) -> str:
        self.resolve_paths()

        if not self.ssr_path:
            # Try to resolve the path dynamically now
            raise ValueError("No SSR path set for this controller")

        # Now that we've built the header, we can remove it from the server data
        # This makes our cache more efficient, since metadata changes don't affect
        # the actual page contents.
        server_data = server_data.model_copy(update={"metadata": None})

        # TODO: Provide a function to automatically sniff for the client view folder
        start = monotonic_ns()
        try:
            ssr_html = render_ssr(
                self.ssr_path.read_text(),
                server_data,
                hard_timeout=self.hard_ssr_timeout,
            )
        except V8RuntimeError as e:
            # Try to parse the file sources and re-raise the error with the
            # maps on the current filesystem
            if self.source_map is not None:
                # parse() is a no-op if the source map is already parsed, so we can do it again
                self.source_map.parse()
                raise V8RuntimeError(self.source_map.map_exception(str(e)))
            raise e

        ssr_duration = (monotonic_ns() - start) / 1e9
        if ssr_duration > self.slow_ssr_threshold:
            LOGGER.warning(f"Slow SSR render detected: {ssr_duration:.2f}s")
        else:
            LOGGER.debug(f"SSR render took {ssr_duration:.2f}s")

        return cast(str, ssr_html)

    def _build_header(self, metadata: Metadata) -> list[str]:
        """
        Builds the header for this controller. Returns the list of tags that will be injected into the
        <head> tag of the rendered page.

        """
        tags: list[str] = []

        def format_optional_keys(payload: Mapping[str, str | bool | None]) -> str:
            attributes: list[str] = []
            for key, value in payload.items():
                if value is None:
                    continue
                elif isinstance(value, bool):
                    # Boolean attributes can just be represented by just their key
                    if value:
                        attributes.append(key)
                    else:
                        continue
                else:
                    attributes.append(f'{key}="{value}"')
            return " ".join(attributes)

        if metadata.title:
            tags.append(f"<title>{metadata.title}</title>")

        for meta_definition in metadata.metas:
            meta_attributes = {
                "name": meta_definition.name,
                "content": meta_definition.content,
                **meta_definition.optional_attributes,
            }
            tags.append(f"<meta {format_optional_keys(meta_attributes)} />")

        for script_definition in metadata.scripts:
            script_attributes: dict[str, str | bool] = {
                "src": script_definition.src,
                "async": script_definition.asynchronous,
                "defer": script_definition.defer,
                **script_definition.optional_attributes,
            }
            tags.append(f"<script {format_optional_keys(script_attributes)}></script>")

        for link_definition in metadata.links:
            link_attributes = {
                "rel": link_definition.rel,
                "href": link_definition.href,
                **link_definition.optional_attributes,
            }
            tags.append(f"<link {format_optional_keys(link_attributes)} />")

        return tags

    def _get_client_functions(self) -> Iterable[tuple[str, Callable, FunctionMetadata]]:
        """
        Returns all of the client-callable functions for this controller. Right now we force
        client accessible functions to either be wrapped by @sideeffect or @passthrough.

        """
        # Iterate over all the functions in this class and see which ones have a _metadata attribute
        # We specifically traverse through the MRO, except the last one (object class)
        for name, func in getmembers(self, predicate=ismethod):
            try:
                metadata = get_function_metadata(func)
                if metadata.action_type in {
                    FunctionActionType.PASSTHROUGH,
                    FunctionActionType.SIDEEFFECT,
                }:
                    yield name, func, metadata
            except AttributeError:
                continue

    def resolve_paths(self, view_base: Path | None = None, force: bool = True) -> bool:
        """
        Typically used internally by the Mountaineer build pipeline. Calling this function
        sets the active `view_base` of the frontend project, which allows us to resolve the
        built javascripts that are required for this controller.

        :return: Whether we have found all necessary files and fully updated the controller state.

        """
        if not force and self.view_base_path is not None:
            return False

        # Try to resolve the view base path from the global config
        if view_base is None:
            try:
                config = get_config()
                if config.PACKAGE:
                    view_base = Path(resolve_package_path(config.PACKAGE)) / "views"
            except (ValueError, PackageNotFoundError):
                # Config isn't registered yet
                pass

        if view_base is None:
            # Unable to resolve, no-op
            return False

        script_name = underscore(self.__class__.__name__)
        self.view_base_path = view_base

        # We'll update this bool if we can't find any dependencies
        found_dependencies = True

        # The SSR path is going to be static
        ssr_path = view_base / "_ssr" / f"{script_name}.js"
        if ssr_path.exists():
            self.ssr_path = ssr_path
            ssr_map_path = ssr_path.with_suffix(".js.map")
            self.source_map = (
                SourceMapParser(ssr_map_path) if ssr_map_path.exists() else None
            )
        else:
            found_dependencies = False

        # Find the md5-converted cache path
        md5_script_pattern = re_compile(script_name + "-" + "[a-f0-9]{32}" + ".js")
        if (view_base / "_static").exists():
            self.bundled_scripts = [
                path.name
                for path in (view_base / "_static").iterdir()
                if md5_script_pattern.match(path.name) and ".js.map" not in path.name
            ]
            if not self.bundled_scripts:
                found_dependencies = False
            LOGGER.debug(
                f"[{self.__class__.__name__}] Resolved paths... {self.bundled_scripts}"
            )
        else:
            found_dependencies = False

        return found_dependencies

    def _merge_metadatas(self, metadatas: list[Metadata]):
        """
        Merges a list of metadata objects, sorted by priority. Some fields will
        take the union (like scripts) - others will prioritize earlier entries (title).

        """
        # Keep track of the unique values we've seen already to ensure that we are:
        # 1. Only including unique values
        # 2. Ranking them in the same order as they were provided
        metas: set[MetaAttribute] = set()
        links: set[LinkAttribute] = set()
        scripts: set[ScriptAttribute] = set()

        final_metadata = Metadata()

        for metadata in metadatas:
            final_metadata.title = final_metadata.title or metadata.title

            final_metadata.metas.extend(
                [element for element in metadata.metas if element not in metas]
            )
            final_metadata.links.extend(
                [element for element in metadata.links if element not in links]
            )
            final_metadata.scripts.extend(
                [element for element in metadata.scripts if element not in scripts]
            )

            metas |= set(metadata.metas)
            links |= set(metadata.links)
            scripts |= set(metadata.scripts)

        return final_metadata



from re import sub as re_sub
from typing import Callable

import pytest
from fastapi import APIRouter
from fastapi.openapi.utils import get_openapi

from mountaineer.client_builder.build_links import OpenAPIToTypescriptLinkConverter


def view_endpoint_regular():
    pass


def view_endpoint_path_params(path_a: str, path_b: int):
    pass


def view_endpoint_query_params(query_a: str, query_b: int | None = None):
    pass


@pytest.mark.parametrize(
    "url, endpoint, expected_link",
    [
        # Regular view endpoint, no render params
        (
            "/regular/",
            view_endpoint_regular,
            (
                """
                export const getLink = ({}:{}) => {
                    const url = `/regular/`;
                    const queryParameters: Record<string,any> = {};
                    const pathParameters: Record<string,any> = {};
                    return __getLink({
                        rawUrl: url,
                        queryParameters,
                        pathParameters
                    });
                };
                """
            ),
        ),
        # View endpoint with path params
        (
            "/url_params/{path_a}/{path_b}",
            view_endpoint_path_params,
            (
                """
                export const getLink = ({
                    path_a,
                    path_b
                }:{
                    path_a: string,
                    path_b: number
                }) => {
                    const url = `/url_params/{path_a}/{path_b}`;
                    const queryParameters: Record<string,any> = {};
                    const pathParameters: Record<string,any> = {
                        path_a,
                        path_b
                    };
                    return __getLink({
                        rawUrl: url,
                        queryParameters,
                        pathParameters
                    });
                };
                """
            ),
        ),
        # View endpoint with query params
        (
            "/query_params",
            view_endpoint_query_params,
            (
                """
                export const getLink = ({
                    query_a,
                    query_b
                }:{
                    query_a: string,
                    query_b?: null | number
                }) => {
                    const url = `/query_params`;
                    const queryParameters: Record<string,any> = {
                        query_a,
                        query_b
                    };
                    const pathParameters: Record<string,any> = {};
                    return __getLink({
                        rawUrl: url,
                        queryParameters,
                        pathParameters
                    });
                };
                """
            ),
        ),
    ],
)
def test_convert(url: str, endpoint: Callable, expected_link: str):
    # Each function needs a response model because we expect that all @sideeffect
    # and @passthrough functions will have an automatically defined response model
    router = APIRouter()
    router.get(url)(endpoint)

    builder = OpenAPIToTypescriptLinkConverter()
    openapi_spec = get_openapi(title="", version="", routes=router.routes)
    link_fn = builder.convert(openapi_spec)

    assert re_sub(r"\s+", "", link_fn) == re_sub(r"\s+", "", expected_link)



from re import sub as re_sub
from time import monotonic_ns
from uuid import UUID, uuid4

import pytest
from pydantic import BaseModel

from mountaineer.__tests__.fixtures import get_fixture_path
from mountaineer.ssr import V8RuntimeError, render_ssr


def test_ssr_speed_baseline():
    all_measurements: list[float] = []

    js_contents = get_fixture_path("home_controller_ssr_with_react.js").read_text()

    class FakeModel(BaseModel):
        # We need to bust the cache
        random_id: UUID

        model_config = {
            "frozen": True,
        }

    for _ in range(50):
        start = monotonic_ns()
        render_ssr(
            js_contents,
            FakeModel(random_id=uuid4()),
            hard_timeout=1,
        )
        all_measurements.append((monotonic_ns() - start) / 1e9)

    assert max(all_measurements) < 0.5


# We expect an exception is raised in our thread so we don't need
# the additional log about it
def test_ssr_timeout():
    js_contents = get_fixture_path("complex_controller_ssr_with_react.js").read_text()

    class FakeWaitDurationModel(BaseModel):
        # Accepts this variable to determine how many ~2s intervals to spend
        # doing synthetic work
        delay_loops: int

        random_id: UUID

        model_config = {
            "frozen": True,
        }

    start = monotonic_ns()
    with pytest.raises(TimeoutError):
        render_ssr(
            script=js_contents,
            render_data=FakeWaitDurationModel(delay_loops=5, random_id=uuid4()),
            hard_timeout=0.5,
        )
    assert ((monotonic_ns() - start) / 1e9) < 1.0


def test_ssr_exception_context():
    """
    Ensure we report the context of V8 runtime exceptions.
    """

    class FakeModel(BaseModel):
        random_id: UUID

        model_config = {
            "frozen": True,
        }

    js_contents = """
    var SSR = {
        x: () => {
            throw new Error('custom_error_text')
        }
    };
    """

    # with pytest.raises(V8RuntimeError, match="custom_error_text"):
    try:
        render_ssr(
            script=js_contents,
            render_data=FakeModel(random_id=uuid4()),
            hard_timeout=0,
        )
    except V8RuntimeError as e:
        assert re_sub(r"\s+", "", str(e)) == (
            re_sub(
                r"\s+",
                "",
                """
            Error calling function 'x': Error: custom_error_text
            Stack: Error: custom_error_text
                    at Object.x (<anonymous>:4:19)
            """,
            )
        )



from mountaineer import AppController, LinkAttribute, Metadata
from mountaineer.js_compiler.postcss import PostCSSBundler

from ci_webapp.config import AppConfig
from ci_webapp.controllers.complex import ComplexController
from ci_webapp.controllers.detail import DetailController
from ci_webapp.controllers.home import HomeController
from ci_webapp.controllers.stream import StreamController

controller = AppController(
    global_metadata=Metadata(
        links=[LinkAttribute(rel="stylesheet", href="/static/app_main.css")]
    ),
    custom_builders=[
        PostCSSBundler(),
    ],
    config=AppConfig(),
)
controller.register(HomeController())
controller.register(DetailController())
controller.register(ComplexController())
controller.register(StreamController())



import subprocess
from itertools import product
from os import environ
from pathlib import Path
from time import sleep
from uuid import uuid4

import pytest
import tomllib
from click import secho
from packaging.requirements import Requirement
from packaging.version import Version
from requests import get

from create_mountaineer_app.__tests__.common import wait_for_database_to_be_ready
from create_mountaineer_app.builder import (
    build_project,
    environment_from_metadata,
    should_copy_path,
)
from create_mountaineer_app.generation import EditorType, ProjectMetadata
from create_mountaineer_app.io import get_free_port
from create_mountaineer_app.templates import get_template_path


@pytest.mark.parametrize(
    "root_path, input_path, expected_copy",
    [
        (
            Path("base"),
            Path("base/myproject/__pycache__/test.pyc"),
            False,
        ),
        (
            Path("base"),
            Path("base/myproject/.git/HEAD"),
            False,
        ),
        (
            Path("base"),
            Path("base/myproject/regular_file.txt"),
            True,
        ),
        (
            # Root paths with hidden files should be excluded from our filtering logic
            # We don't control where installers like pipx place our library
            Path(".cache/pipx/venvs"),
            Path(".cache/pipx/venvs/myproject/regular_file.txt"),
            True,
        ),
    ],
)
def test_copy_path(root_path: Path, input_path: Path, expected_copy: bool):
    assert should_copy_path(root_path, input_path) == expected_copy


@pytest.mark.parametrize(
    "use_poetry, use_tailwind, editor_config, create_stub_files",
    list(
        product(
            # Use poetry
            [False, True],
            # Use tailwind
            [False, True],
            # Editor config
            [None, EditorType.VSCODE, EditorType.VIM, EditorType.ZED],
            # Create stub files
            [False, True],
        )
    ),
)
@pytest.mark.integration_tests
def test_valid_permutations(
    tmpdir: str,
    use_poetry: bool,
    use_tailwind: bool,
    editor_config: EditorType | None,
    create_stub_files: bool,
):
    """
    Ensures that regardless of the input parameters
    used to generate our new project, it will successfully
    run and return the expected endpoints.

    """
    new_project_dir = Path(tmpdir) / str(uuid4())
    new_project_dir.mkdir()

    # Assume the create-mountaineer-app project is in the same directory as the
    # main mountaineer package. We use a slight hack here assuming that our file
    # directory is oriented like "mountaineer/create_mountaineer_app/create_mountaineer_app/templates"
    # main package.
    main_mountaineer_path = get_template_path("../../../").resolve()

    # Verify this is pointing to where we expect
    if not (main_mountaineer_path / "pyproject.toml").exists():
        raise ValueError("Unable to find the main mountaineer package.")

    app_test_port = get_free_port()
    postgres_port = get_free_port()
    secho(f"Found free port for test server: {app_test_port}", fg="green")

    metadata = ProjectMetadata(
        project_name="my_project",
        author_name="John Appleseed",
        author_email="test@email.com",
        project_path=new_project_dir,
        use_poetry=use_poetry,
        use_tailwind=use_tailwind,
        editor_config=editor_config,
        create_stub_files=create_stub_files,
        postgres_port=postgres_port,
        # Stub, not used in template generation since we also have
        # the mountaineer_dev_path
        mountaineer_min_version="0.1.0",
        mountaineer_dev_path=main_mountaineer_path,
    )

    build_project(metadata)

    # Launch docker to host the default database
    docker_compose_env = {
        **environ,
        "COMPOSE_PROJECT_NAME": f"test_project-{uuid4()}",
    }
    subprocess.run(
        ["docker-compose", "up", "-d"],
        cwd=metadata.project_path,
        check=True,
        env=docker_compose_env,
    )

    # Wait until the database is ready
    wait_for_database_to_be_ready(metadata)

    environment = environment_from_metadata(metadata)

    # Make sure the required models are created
    create_db_process = environment.run_command(["createdb"], metadata.project_path)
    output, errors = create_db_process.communicate()
    if create_db_process.returncode != 0:
        secho(output.decode("utf-8"), fg="red")
        secho(errors.decode("utf-8"), fg="red")
        raise ValueError("Failed to create database.")

    # Make sure we can build the files without any errors
    build_process = environment.run_command(["build"], metadata.project_path)
    output, errors = build_process.communicate()
    if build_process.returncode != 0:
        secho(output.decode("utf-8"), fg="red")
        secho(errors.decode("utf-8"), fg="red")
        raise ValueError("Failed to build project.")

    # Now launch the server in the background
    process = environment.run_command(
        ["runserver", "--port", str(app_test_port)], metadata.project_path
    )

    try:
        # Wait up to 10s for the server to start and be accessible
        max_wait = 10
        while max_wait > 0:
            try:
                response = get(f"http://localhost:{app_test_port}")
                if create_stub_files:
                    if response.ok:
                        break
                else:
                    if response.status_code == 404:
                        break
            except Exception:
                pass
            secho(f"Waiting for server to start (remaining: {max_wait})...")
            sleep(1)
            max_wait -= 1

        assert max_wait > 0, "Server start timed out."

        secho("Test server started successfully", fg="green")

        # Perform the fetch tests
        response = get(f"http://localhost:{app_test_port}")
        if create_stub_files:
            assert response.ok
        else:
            assert response.status_code == 404

        response = get(f"http://localhost:{app_test_port}/not_found")
        assert not response.ok
    finally:
        if process.returncode is None:
            process.terminate()
            process.wait()
            secho("Server shut down...")
        else:
            secho(f"Server exited with code {process.returncode}")

        secho("Shutting down docker...")
        subprocess.run(
            ["docker-compose", "down"],
            cwd=metadata.project_path,
            check=True,
            env=docker_compose_env,
        )
        secho("Docker shut down successfully.")


@pytest.mark.parametrize(
    "use_poetry",
    [False, True],
)
def test_build_version_number(use_poetry: bool, tmp_path: Path):
    metadata = ProjectMetadata(
        project_name="my_project",
        author_name="John Appleseed",
        author_email="test@email.com",
        project_path=tmp_path,
        use_poetry=use_poetry,
        use_tailwind=False,
        editor_config=None,
        create_stub_files=False,
        # Stub, not used in template generation since we also have
        # the mountaineer_dev_path
        mountaineer_min_version="0.2.5",
        mountaineer_dev_path=None,
    )
    build_project(metadata, install_deps=False)

    pyproject_contents = (tmp_path / "pyproject.toml").read_text()
    package_requirements = tomllib.loads(pyproject_contents)

    if use_poetry:
        # Parse the poetry convention into the PEP 508 specifier format
        raw_version = package_requirements["tool"]["poetry"]["dependencies"][
            "mountaineer"
        ]
        assert raw_version == "^0.2.5"
        version = f"mountaineer {raw_version}".replace("^", ">=").replace("~", "~=")
    else:
        version = package_requirements["project"]["dependencies"][0]

    req = Requirement(version)
    assert req.specifier.contains(Version("0.2.5"))



from click import command, option
from mountaineer.cli import handle_runserver, handle_watch, handle_build
from mountaineer.database.cli import handle_createdb
from mountaineer.io import async_to_sync

from {{project_name}} import models
from {{project_name}}.config import AppConfig


@command()
@option("--port", default=5006, help="Port to run the server on")
def runserver(port: int):
    handle_runserver(
        package="{{project_name}}",
        webservice="{{project_name}}.main:app",
        webcontroller="{{project_name}}.app:controller",
        port=port,
    )


@command()
def watch():
    handle_watch(
        package="{{project_name}}",
        webcontroller="{{project_name}}.app:controller",
    )


@command()
def build():
    handle_build(
        webcontroller="{{project_name}}.app:controller",
    )


@command()
@async_to_sync
async def createdb():
    _ = AppConfig() # type: ignore

    await handle_createdb(models)



from dataclasses import asdict, replace
from json import dumps as json_dumps, loads as json_loads
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from mountaineer.app import AppController
from mountaineer.client_builder.builder import ClientBuilder
from mountaineer.controller import ControllerBase


class ExampleHomeController(ControllerBase):
    url = "/"
    view_path = "/page.tsx"

    def render(self) -> None:
        return None


class ExampleDetailController(ControllerBase):
    url = "/detail/{detail_id}/"
    view_path = "/detail/page.tsx"

    def render(self) -> None:
        return None


@pytest.fixture
def home_controller():
    return ExampleHomeController()


@pytest.fixture
def detail_controller():
    return ExampleDetailController()


@pytest.fixture(scope="function")
def simple_app_controller(
    home_controller: ExampleHomeController, detail_controller: ExampleDetailController
):
    with TemporaryDirectory() as temp_dir_name:
        temp_view_path = Path(temp_dir_name)
        (temp_view_path / "detail").mkdir()

        # Simple view files
        (temp_view_path / "page.tsx").write_text("")
        (temp_view_path / "detail" / "page.tsx").write_text("")

        app_controller = AppController(view_root=temp_view_path)
        app_controller.register(home_controller)
        app_controller.register(detail_controller)
        yield app_controller


@pytest.fixture
def builder(simple_app_controller: AppController):
    return ClientBuilder(simple_app_controller)


def test_generate_static_files(builder: ClientBuilder):
    builder.generate_static_files()


def test_generate_model_definitions(builder: ClientBuilder):
    builder.generate_model_definitions()


def test_generate_action_definitions(builder: ClientBuilder):
    builder.generate_action_definitions()


def test_generate_view_definitions(builder: ClientBuilder):
    builder.generate_link_shortcuts()


def test_generate_link_aggregator(builder: ClientBuilder):
    builder.generate_link_aggregator()


def test_generate_view_servers(builder: ClientBuilder):
    builder.generate_view_servers()


def test_generate_index_file(builder: ClientBuilder):
    builder.generate_index_file()


def test_cache_is_outdated_no_cache(builder: ClientBuilder):
    # No cache
    builder.build_cache = None
    assert builder.cache_is_outdated() is True


def test_cache_is_outdated_no_existing_data(builder: ClientBuilder, tmp_path: Path):
    builder.build_cache = tmp_path

    assert builder.cache_is_outdated() is True

    # Ensure that we've written to the cache
    cache_path = tmp_path / "client_builder_openapi.json"
    assert cache_path.exists()
    assert set(json_loads(cache_path.read_text()).keys()) == {
        "ExampleHomeController",
        "ExampleDetailController",
    }


def test_cache_is_outdated_existing_data(
    builder: ClientBuilder,
    tmp_path: Path,
    home_controller: ExampleHomeController,
    detail_controller: ExampleDetailController,
):
    builder.build_cache = tmp_path

    # Ensure that we've written to the cache
    cache_path = tmp_path / "client_builder_openapi.json"
    cache_path.write_text(
        json_dumps(
            {
                "ExampleHomeController": {
                    "action": builder.openapi_action_specs[home_controller],
                    "render": asdict(builder.openapi_render_specs[home_controller]),
                },
                "ExampleDetailController": {
                    "action": builder.openapi_action_specs[detail_controller],
                    "render": asdict(builder.openapi_render_specs[detail_controller]),
                },
            },
            sort_keys=True,
        )
    )

    assert builder.cache_is_outdated() is False


def test_cache_is_outdated_url_change(
    builder: ClientBuilder,
    tmp_path: Path,
    home_controller: ExampleHomeController,
    detail_controller: ExampleDetailController,
):
    builder.build_cache = tmp_path

    cache_path = tmp_path / "client_builder_openapi.json"
    cache_path.write_text(
        json_dumps(
            {
                "ExampleHomeController": {
                    "action": builder.openapi_action_specs[home_controller],
                    "render": asdict(
                        # Only modify the render attribute. Simulate a user changing the URL
                        # of a component, which does require a FE rebuild.
                        replace(
                            builder.openapi_render_specs[home_controller],
                            url="/new_url",
                        )
                    ),
                },
                "ExampleDetailController": {
                    "action": builder.openapi_action_specs[detail_controller],
                    "render": asdict(builder.openapi_render_specs[detail_controller]),
                },
            },
            sort_keys=True,
        )
    )

    assert builder.cache_is_outdated() is True



from fastapi import Depends
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.pool import AsyncAdaptedQueuePool, NullPool

from mountaineer.database.config import DatabaseConfig, PoolType
from mountaineer.dependencies import CoreDependencies
from mountaineer.logging import LOGGER

# We share the connection pool across the entire process
GLOBAL_ENGINE: dict[str, AsyncEngine] = {}


def engine_from_config(config: DatabaseConfig, force_new: bool = False):
    global GLOBAL_ENGINE

    if not config.SQLALCHEMY_DATABASE_URI:
        raise RuntimeError(f"No SQLALCHEMY_DATABASE_URI set: {config}")

    if force_new or str(config.SQLALCHEMY_DATABASE_URI) not in GLOBAL_ENGINE:
        GLOBAL_ENGINE[str(config.SQLALCHEMY_DATABASE_URI)] = create_async_engine(
            str(config.SQLALCHEMY_DATABASE_URI),
            poolclass=(
                NullPool
                if config.DATABASE_POOL_TYPE == PoolType.NULL
                else AsyncAdaptedQueuePool
            ),
        )

    return GLOBAL_ENGINE[str(config.SQLALCHEMY_DATABASE_URI)]


async def get_db(
    config: DatabaseConfig = Depends(
        CoreDependencies.get_config_with_type(DatabaseConfig)
    ),
):
    """
    Gets the SQLAlchemy engine registered for your application. Since our
    DatabaseConfig specifies global parameters, by default this engine is
    shared across the whole application.

    If called via dependency injection, which is the most common case,
    we will automatically resolve the config for you. Clients just call
    the main function, which will resolve sub-dependencies.

    ```python
    from sqlalchemy.ext.asyncio import AsyncEngine

    async def render(
        self,
        engine: AsyncEngine = Depends(DatabaseDependencies.get_db)
    ):
        ...
    ```

    """
    return engine_from_config(config)


async def get_db_session(
    engine: AsyncEngine = Depends(get_db),
):
    """
    Gets a new, limited scope async session from the global engine. Anything
    that occurs within this managed scope is wrapped in a transaction, so you
    must manually commit your changes with `await session.commit()` when finished.

    Like `get_db`, the engine will be resolved automatically if you call this
    within a dependency injection context.


    ```python
    from sqlalchemy.ext.asyncio import AsyncSession

    async def render(
        self,
        db_session: AsyncSession = Depends(DatabaseDependencies.get_db_session)
    ):
        ...
    ```

    """
    session_maker = async_sessionmaker(engine, expire_on_commit=False)
    async with session_maker() as session:
        try:
            yield session
        except Exception as e:
            # SQLAlchemy provides rollback support automatically with the async session manager
            LOGGER.exception(
                f"Error in user code, rolling back uncommitted db changes: {e}"
            )
            raise


async def unregister_global_engine():
    global GLOBAL_ENGINE
    if GLOBAL_ENGINE is not None:
        for engine in GLOBAL_ENGINE.values():
            await engine.dispose()
        GLOBAL_ENGINE = {}



from importlib.resources import as_file, files
from pathlib import Path


def get_core_view_path(asset_name: str) -> Path:
    with as_file(files(__name__).joinpath(asset_name)) as path:
        return Path(path)



from inspect import signature

import pytest
from fastapi import Depends, Request
from typing_extensions import Callable

from mountaineer.dependencies.base import (
    DependenciesBase,
    get_function_dependencies,
    isolate_dependency_only_function,
)


@pytest.mark.asyncio
async def test_get_function_dependencies_recursive():
    async def dep_1():
        return 1

    async def dep_2(
        dep_1: int = Depends(dep_1),
    ):
        return 2 + dep_1

    def dep_3(dep_2: int = Depends(dep_2)):
        return dep_2 + 3

    with pytest.warns(DeprecationWarning):

        class ExampleDependencies(DependenciesBase):
            dep_1: Callable
            dep_2: Callable
            dep_3: Callable

    ExampleDependencies.dep_1 = dep_1
    ExampleDependencies.dep_2 = dep_2
    ExampleDependencies.dep_3 = dep_3

    async with get_function_dependencies(callable=ExampleDependencies.dep_3) as values:
        result = ExampleDependencies.dep_3(**values)
        assert result == 6


def test_incorrect_static_method():
    """
    Ensure static methods will throw an error on init

    """
    with pytest.warns(DeprecationWarning), pytest.raises(TypeError):

        class ExampleIncorrectDependency(DependenciesBase):
            @staticmethod
            async def dep_1():
                return 1


@pytest.mark.asyncio
async def test_dependency_overrides():
    def dep_1():
        return "Original Value"

    def dep_2(
        dep_1: str = Depends(dep_1),
    ):
        return f"Final Value: {dep_1}"

    def mocked_dep_1():
        return "Mocked Value"

    async with get_function_dependencies(
        callable=dep_2,
        dependency_overrides={
            dep_1: mocked_dep_1,
        },
    ) as values:
        result = dep_2(**values)
        assert result == "Final Value: Mocked Value"


class ExamplePayload:
    value: int


@pytest.mark.asyncio
async def test_isolate_dependency_only_function():
    def test_dependency():
        return 1

    def test_complex_function(
        payload: ExamplePayload,
        request: Request,
        resolved_dep: int = Depends(test_dependency),
    ):
        return resolved_dep

    modified_function = isolate_dependency_only_function(
        test_complex_function,
    )
    new_signature = signature(modified_function)
    assert set(new_signature.parameters.keys()) == {"resolved_dep"}

    # Now try to execute the dependencies
    async with get_function_dependencies(
        callable=modified_function,
    ) as values:
        assert values == {"resolved_dep": 1}



import asyncio
from threading import Thread

from fastapi import FastAPI
from uvicorn import Config
from uvicorn.server import Server


class UvicornThread(Thread):
    def __init__(self, *, app: FastAPI, port: int, log_level: str = "info"):
        super().__init__(daemon=True)
        self.app = app
        self.port = port
        self.log_level = log_level
        self.server: Server | None = None

    def run(self):
        loop = asyncio.new_event_loop()
        config = Config(
            app=self.app,
            port=self.port,
            reload=False,
            access_log=False,
            loop="asyncio",
            log_level=self.log_level,
        )
        self.server = Server(config)
        loop.run_until_complete(self.server.serve())

    def stop(self):
        if self.server is not None:
            self.server.should_exit = True



from enum import Enum
from typing import Any

from pydantic import PostgresDsn, model_validator
from pydantic_core import PydanticUndefinedType
from pydantic_settings import BaseSettings


class PoolType(Enum):
    # Default value, no in-memory pooling at the SQLAlchemy layer. This assumes
    # you'll establish a 3rd party connection pool closer to the database
    # layer - or you have low enough traffic that you don't need one at all.
    NULL = "NULL"

    # Fixed quantity for each process spawned
    # This corresponds to the SQLAlchemy `AsyncAdaptedQueuePool`
    FIXED_PROCESS = "FIXED_PROCESS"


class DatabaseConfig(BaseSettings):
    POSTGRES_HOST: str
    POSTGRES_USER: str
    POSTGRES_PASSWORD: str
    POSTGRES_DB: str
    POSTGRES_PORT: int = 5432
    SQLALCHEMY_DATABASE_URI: PostgresDsn | None = None

    DATABASE_POOL_TYPE: PoolType = PoolType.NULL

    @model_validator(mode="before")
    def build_db_connection(cls, values: Any) -> Any:
        # Support users specifying some defaults via the pydantic
        # model directly
        cls_defaults = {
            field_name: field.default
            for field_name, field in cls.model_fields.items()
            if field.default and not isinstance(field.default, PydanticUndefinedType)
        }

        all_values = {**cls_defaults, **values}

        if not values.get("SQLALCHEMY_DATABASE_URI"):
            values["SQLALCHEMY_DATABASE_URI"] = PostgresDsn.build(  # type: ignore
                scheme="postgresql+asyncpg",
                username=all_values["POSTGRES_USER"],
                password=all_values["POSTGRES_PASSWORD"],
                host=all_values["POSTGRES_HOST"],
                port=int(all_values.get("POSTGRES_PORT", 5432)),
                path=all_values["POSTGRES_DB"],
            )

        return values



"""
Utilities for client unit and integration tests
"""
from functools import wraps
from inspect import isawaitable, iscoroutinefunction, signature
from tempfile import NamedTemporaryFile
from time import monotonic_ns
from typing import Any

import pytest
from sqlalchemy import text
from sqlalchemy.exc import DBAPIError
from sqlalchemy.ext.asyncio import AsyncEngine

from mountaineer.database import SQLModel
from mountaineer.logging import LOGGER


class ExecutionTooLong(Exception):
    pass


def benchmark_function(
    max_time_seconds: int | float, time_budget_seconds: int | float = 5
):
    """
    Wrap test functions in a timer that will enforce that the core logic completes
    in less than `max_time_seconds` seconds. Injects `start_timing` and `end_timing` into
    the function kwargs that will be called. Client callers should call these to scope
    where they want us to measure the core logic.

    :param time_budget_seconds: How many seconds we have to run benchmarking. This function will run
      at least 1 benchmark always, but the time budget allows users to better calibrate benchmark duration
      by taking the average of multiple runs.

    """
    import pyinstrument

    def wrapper_fn(test_func):
        # We want to remove our custom functions from the signature, since pytest will natively
        # try to inject fixtures in this place
        orig_sig = signature(test_func)
        new_params = [
            p
            for name, p in orig_sig.parameters.items()
            if name not in {"start_timing", "end_timing"}
        ]
        new_sig = orig_sig.replace(parameters=new_params)

        # Make sure that our function is a coroutine (if it is we assume user has decorated it with
        # pytest.mark.asyncio), otherwise we should raise an error because pytest won't know how
        # to run it
        if not iscoroutinefunction(test_func):
            raise Exception(
                f"Test function {test_func.__name__} is not a coroutine function. Please decorate it with pytest.mark.asyncio"
            )

        async def single_time_test(fn, *args, **kwargs):
            # Try to run the test function regularly, and time it
            # Instrumenting profilers typically will slow down execution time so we want
            # to take time of the raw, non-instrumented function for benchmarking
            start: float | None = None
            end: float | None = None

            def start_timing():
                nonlocal start
                start = monotonic_ns()

            def end_timing():
                nonlocal end
                end = monotonic_ns()

            result = test_func(
                *args, **kwargs, start_timing=start_timing, end_timing=end_timing
            )
            if isawaitable(result):
                result = await result

            return (start, end, result)

        @wraps(test_func)
        async def wrapper(*args, **kwargs):
            bound = new_sig.bind(*args, **kwargs)
            bound.apply_defaults()

            average_duration: float | None = None

            try:
                timed_durations: list[tuple[int, int]] = []
                results: list[Any] = []
                global_start = monotonic_ns()

                while monotonic_ns() - global_start < time_budget_seconds * 1e9:
                    start, end, result = await single_time_test(
                        test_func, *args, **kwargs
                    )

                    if start is None:
                        raise Exception("Test function did not call start_timing")
                    if end is None:
                        raise Exception("Test function did not call end_timing")

                    timed_durations.append((start, end))
                    results.append(result)

                average_duration = sum(
                    (end - start) for start, end in timed_durations
                ) / len(timed_durations)

                LOGGER.info(
                    f"Collected {len(results)} timed durations in {(monotonic_ns() - global_start) / 1e9}"
                )
                LOGGER.info(f"Test function took average: {average_duration / 1e9}")

                if average_duration / 1e9 > max_time_seconds:
                    raise ExecutionTooLong()

                return results[0]

            except ExecutionTooLong as e:
                LOGGER.error(f"Test function failed due to: {e}")

                # This should already be true, but we want to be explicit to help mypy
                assert average_duration is not None

                profiler = pyinstrument.Profiler()
                output_filename: str | None = None

                def start_timing():
                    nonlocal profiler
                    profiler.start()

                def end_timing():
                    nonlocal profiler
                    nonlocal output_filename
                    profiler.stop()
                    with NamedTemporaryFile(delete=False, suffix=".html") as file:
                        file.write(profiler.output_html().encode())
                    LOGGER.warning(f"PyInstrument profile saved at: {file.name}")
                    output_filename = file.name

                result = test_func(
                    *args, **kwargs, start_timing=start_timing, end_timing=end_timing
                )
                if isawaitable(result):
                    await result

                pytest.fail(
                    f"Test function failed in {average_duration/1e9}s and profiles generated; Pyinstrument: {output_filename}"
                )

        wrapper.__signature__ = new_sig  # type: ignore
        return wrapper

    return wrapper_fn


async def bootstrap_database(engine: AsyncEngine):
    # Test if we can connect to the database, if not throw an early error since
    # the user likely hasn't booted up the test database
    try:
        async with engine.connect() as conn:
            # Optionally, you can perform a simple query to ensure the connection is valid
            await conn.execute(text("SELECT 1"))
    except (DBAPIError, OSError) as e:
        # Handle specific database connection errors or raise a custom exception
        raise RuntimeError(
            "Failed to connect to the test database. Please ensure the database is running and accessible."
        ) from e

    # We do this at the beginning (instead of auto-cleaning at the end) to make sure
    # that interrupted runs aren't polluting our database state for a new run, and to
    # allow us to inspect failed tests after they're run
    LOGGER.info("Cleaning up test tables")
    async with engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.drop_all)

    async with engine.begin() as conn:
        await conn.run_sync(SQLModel.metadata.create_all)

        # Log the tables that were created
        result = await conn.execute(
            text(
                """
            SELECT table_name
            FROM information_schema.tables
            WHERE table_schema = 'public'
        """
            )
        )
        tables = "\n".join([f"* {table[0]}" for table in result.fetchall()])
        LOGGER.info(f"Created tables:\n{tables}")



import pytest

from mountaineer.js_compiler.postcss import PostCSSBundler
from mountaineer.paths import ManagedViewPath


@pytest.mark.parametrize(
    "view_path, expected_output_name",
    [
        (
            "root_style.css",
            "root_style.css",
        ),
        (
            "root_style.scss",
            "root_style.css",
        ),
        (
            "home/style.css",
            "home_style.css",
        ),
        (
            "home/nested/other/style.css",
            "home_nested_other_style.css",
        ),
    ],
)
def test_get_style_output_name(view_path: str, expected_output_name: str):
    view_root = ManagedViewPath.from_view_root("fake-root")

    bundler = PostCSSBundler()
    assert bundler.get_style_output_name(view_root / view_path) == expected_output_name



from typing import (
    AbstractSet,
    Any,
    Dict,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass
from pydantic_core import (
    PydanticUndefined as Undefined,
    PydanticUndefinedType as UndefinedType,
)
from sqlalchemy import Column
from sqlmodel._compat import (
    finish_init,
    post_init_field_info,
    sqlmodel_init,
)
from sqlmodel.main import (
    FieldInfo,
    NoArgAnyCallable,
    SQLModel as SQLModelBase,
    SQLModelMetaclass as SQLModelMetaclassBase,
)
from typing_extensions import dataclass_transform


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: Optional[bool] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
    max_digits: Optional[int] = None,
    decimal_places: Optional[int] = None,
    min_items: Optional[int] = None,
    max_items: Optional[int] = None,
    unique_items: Optional[bool] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    allow_mutation: bool = True,
    regex: Optional[str] = None,
    discriminator: Optional[str] = None,
    repr: bool = True,
    primary_key: Union[bool, UndefinedType] = Undefined,
    foreign_key: Any = Undefined,
    unique: Union[bool, UndefinedType] = Undefined,
    nullable: Union[bool, UndefinedType] = Undefined,
    index: Union[bool, UndefinedType] = Undefined,
    sa_type: Union[Type[Any], Any, UndefinedType] = Undefined,
    sa_column: Union[Column, UndefinedType] = Undefined,  # type: ignore
    sa_column_args: Union[Sequence[Any], UndefinedType] = Undefined,
    sa_column_kwargs: Union[Mapping[str, Any], UndefinedType] = Undefined,
    schema_extra: Optional[dict[str, Any]] = None,
) -> Any:
    """
    Allow instantiated `sa_type` to be used. This permits DateTime(timezone=True) instead
    of just passing a vanilla DateTime. Since these are passed through directly to the column
    syntax this is the same as a regular column declaration:

    https://github.com/tiangolo/sqlmodel/blob/6b562358fc1e857dd1ce4b8b23a9f68c0337430d/sqlmodel/main.py#L561

    Original: https://github.com/tiangolo/sqlmodel/blob/6b562358fc1e857dd1ce4b8b23a9f68c0337430d/sqlmodel/main.py

    """
    current_schema_extra = schema_extra or {}
    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        max_digits=max_digits,
        decimal_places=decimal_places,
        min_items=min_items,
        max_items=max_items,
        unique_items=unique_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        discriminator=discriminator,
        repr=repr,
        primary_key=primary_key,
        foreign_key=foreign_key,
        unique=unique,
        nullable=nullable,
        index=index,
        sa_type=sa_type,
        sa_column=sa_column,
        sa_column_args=sa_column_args,
        sa_column_kwargs=sa_column_kwargs,
        **current_schema_extra,
    )
    post_init_field_info(field_info)
    return field_info


@dataclass_transform(kw_only_default=True, field_specifiers=(Field, FieldInfo))
class GenericSQLModelMetaclass(SQLModelMetaclassBase):
    def __new__(
        cls,
        name: str,
        bases: tuple[Type[Any], ...],
        class_dict: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        """
        Unlike the base SQLModel metaclass, we first fully instantiate the pydantic
        model before transfering the fields to SQLAlchemy. Pydantic V2 has helpful hooks
        within metadata construction that fullfill generic definitions that will be missed
        if we try to naively inspect the type annotations.

        """
        # SQLModel by default won't inherit FieldInfo definitions from parent classes, which is
        # unideal since we want to set it once in a parent class and pass through these values
        # into the children
        # Traverse the base classes to find their Field definitions. We can't grab these
        # directly from the pydantic model, since we want the raw SQLModel attributes
        # instead of the regular pydantic Fields
        parent_field_infos: dict[str, FieldInfo] = {}
        for base in reversed(bases):  # Reverse to ensure correct order of precedence
            if hasattr(base, "_explicit_fieldinfo"):
                parent_field_infos = {**parent_field_infos, **base._explicit_fieldinfo}

        self_field_infos = {
            key: value for key, value in class_dict.items() if not key.startswith("__")
        }

        class_dict = {
            **parent_field_infos,
            **class_dict,
        }

        pydantic_model = cls._validate_pydantic_model(name, bases, class_dict, **kwargs)
        pydantic_annotations = {
            key: field.annotation for key, field in pydantic_model.model_fields.items()
        }

        # Override with the parsed annotations that are extracted from Pydantic
        # This helps typehint the SQLModel class with the correct, resolved types
        class_dict["__annotations__"] = {
            **class_dict.get("__annotations__", {}),
            **pydantic_annotations,
        }

        cls = super().__new__(cls, name, bases, class_dict, **kwargs)

        cls._original_model = pydantic_model
        cls._explicit_fieldinfo = {
            **parent_field_infos,
            **self_field_infos,
        }

        return cls

    @classmethod
    def _validate_pydantic_model(
        cls,
        name: str,
        bases: tuple[Type[Any], ...],
        class_dict: dict[str, Any],
        **kwargs,
    ):
        """
        Given this class declaration, build up a Pydantic model with the type annotation hierarchy
        from base classes. This will resolve the TypeVars and other generic types that are not
        otherwise handled by SQLModel.

        """
        pydantic_bases = [
            base._original_model for base in bases if hasattr(base, "_original_model")
        ]
        parent_resolved_fields = {
            key: value.annotation
            for model in pydantic_bases
            for key, value in model.model_fields.items()
            # Don't pass through TypeVars, we only want to pass concrete types
            if not isinstance(value, TypeVar)
        }
        pydantic_class_dict = {
            **class_dict,
            "__annotations__": {
                **parent_resolved_fields,
                # Allow overrides for this subclass
                **class_dict.get("__annotations__", {}),
            },
        }

        return cast(
            BaseModel,
            ModelMetaclass.__new__(
                ModelMetaclass,
                name,
                (tuple(pydantic_bases) or (BaseModel,)),
                pydantic_class_dict,
                **{key: value for key, value in kwargs.items() if key not in {"table"}},
            ),
        )


class SQLModel(SQLModelBase, metaclass=GenericSQLModelMetaclass):
    def __init__(self, **data: Any):
        if hasattr(self, "_original_model"):
            self._original_model(**data)
        if finish_init.get():
            sqlmodel_init(self=self, data=data)



from contextlib import asynccontextmanager
from pathlib import Path
from time import monotonic_ns
from unittest.mock import patch

import pytest
from fastapi import Depends, Request
from fastapi.testclient import TestClient
from pydantic.main import BaseModel
from starlette.datastructures import Headers

from mountaineer.__tests__.common import calculate_primes
from mountaineer.actions.fields import FunctionActionType, get_function_metadata
from mountaineer.actions.sideeffect import sideeffect
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.app import AppController
from mountaineer.controller import ControllerBase
from mountaineer.logging import LOGGER
from mountaineer.render import RenderBase


class ExampleRenderModel(RenderBase):
    value_a: str
    value_b: str


def test_markup_sideeffect():
    """
    Check that the @sideeffect decorator extracts the expected
    data from our model definition.
    """

    class ExamplePassthroughModel(BaseModel):
        first_name: str

    class TestController(ControllerBase):
        # We declare as complicated a payload as @sideeffect supports so we can
        # see the full amount of metadata properties that are set
        @sideeffect(
            reload=tuple([ExampleRenderModel.value_a]),
        )
        def sideeffect_and_return_data(self) -> ExamplePassthroughModel:
            return ExamplePassthroughModel(
                first_name="John",
            )

    metadata = get_function_metadata(TestController.sideeffect_and_return_data)
    assert metadata.action_type == FunctionActionType.SIDEEFFECT
    assert metadata.get_passthrough_model() == ExamplePassthroughModel
    assert metadata.function_name == "sideeffect_and_return_data"
    assert metadata.reload_states == tuple([ExampleRenderModel.value_a])
    assert isinstance(metadata.render_model, MountaineerUnsetValue)
    assert isinstance(metadata.return_model, MountaineerUnsetValue)


class ControllerCommon(ControllerBase):
    url: str = "/test/{query_id}/"

    def __init__(self):
        super().__init__()
        self.counter = 0
        self.render_counts = 0

    @sideeffect
    def call_sideeffect(self, payload: dict) -> None:
        self.counter += 1

    @sideeffect
    async def call_sideeffect_async(self, payload: dict) -> None:
        self.counter += 1


async def call_sideeffect_common(controller: ControllerCommon):
    app = AppController(view_root=Path())
    app.register(controller)

    @asynccontextmanager
    async def mock_get_render_parameters(*args, **kwargs):
        yield {
            "query_id": 1,
        }

    # After our wrapper is called, our function is now async
    # Avoid the dependency resolution logic since that's tested separately
    with patch(
        "mountaineer.actions.sideeffect.get_render_parameters"
    ) as patched_get_render_params:
        patched_get_render_params.side_effect = mock_get_render_parameters

        # Even if the "request" is not required by our sideeffects, it's required
        # by the function injected by the sideeffect decorator.
        return_value_sync = await controller.call_sideeffect(
            {},
            request=Request({"type": "http"}),  # type: ignore
        )

        return_value_async = await controller.call_sideeffect_async(
            {},
            request=Request({"type": "http"}),  # type: ignore
        )

        # The response payload should be the same both both sync and async endpoints
        expected_response = {
            "sideeffect": ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            ),
            "passthrough": None,
        }

        assert return_value_sync == expected_response
        assert return_value_async == expected_response

        assert controller.counter == 2
        assert controller.render_counts == 2


@pytest.mark.asyncio
async def test_can_call_sideeffect():
    """
    Ensure that we can call the sideeffect, which will in turn
    call the render function to get fresh data.
    """

    class ExampleController(ControllerCommon):
        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_can_call_sideeffect_async_render():
    """
    Render functions can also work asynchronously.
    """

    class ExampleController(ControllerCommon):
        async def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            self.render_counts += 1
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    await call_sideeffect_common(ExampleController())


@pytest.mark.asyncio
async def test_get_render_parameters():
    """
    Given a controller, reproduce the logic of FastAPI to sniff the render()
    function for dependencies and resolve them.

    """
    from mountaineer.actions.sideeffect import get_render_parameters

    found_cookie = None

    def grab_cookie_dependency(request: Request):
        nonlocal found_cookie
        found_cookie = request.cookies.get("test-cookie")
        return found_cookie

    class TestController(ControllerBase):
        url: str = "/test/{query_id}/"

        def render(
            self,
            query_id: int,
            cookie_dependency: str = Depends(grab_cookie_dependency),
        ) -> ExampleRenderModel:
            return ExampleRenderModel(
                value_a="Hello",
                value_b="World",
            )

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = TestController()
    app.register(controller)

    fake_request = Request(
        {
            "type": "http",
            "headers": Headers(
                {
                    "cookie": "test-cookie=cookie-value",
                    # Its important the referer aligns with the controller url, since that is expected
                    # to be the original view page that is calling this sub-function
                    "referer": "http://example.com/test/5/",
                }
            ).raw,
            "http_version": "1.1",
            "scheme": "",
            "client": "",
            "server": "",
            # The URL and method should both be different, to test whether we are able
            # to map the request to the correct endpoint
            "method": "POST",
            "url": "http://localhost/related_action_endpoint",
        }
    )

    async with get_render_parameters(controller, fake_request) as resolved_dependencies:
        assert resolved_dependencies == {
            "cookie_dependency": "cookie-value",
            "query_id": 5,
        }


@pytest.mark.parametrize(
    "use_experimental,min_time,max_time",
    [
        (False, 1, None),
        (True, None, 0.10),
    ],
)
def test_limit_codepath_experimental(
    use_experimental: bool,
    min_time: float | None,
    max_time: float | None,
):
    class ExampleController(ControllerBase):
        url: str = "/test/{query_id}/"

        def render(
            self,
            query_id: int,
        ) -> ExampleRenderModel:
            a = calculate_primes(10000)
            b = calculate_primes(1000000)
            return ExampleRenderModel(
                value_a=f"Hello {a}",
                value_b=f"World {b}",
            )

        @sideeffect(
            reload=(ExampleRenderModel.value_a,),
            experimental_render_reload=use_experimental,
        )
        def call_sideeffect(self, payload: dict) -> None:
            pass

    # We need to load this test controller to an actual application runtime
    # or else we don't have the render() metadata added
    app = AppController(view_root=Path())
    controller = ExampleController()
    app.register(controller)

    controller_definition = app.definition_for_controller(controller)
    sideeffect_url = controller_definition.get_url_for_metadata(
        get_function_metadata(ExampleController.call_sideeffect)
    )

    client = TestClient(app.app)
    start = monotonic_ns()
    response = client.post(
        sideeffect_url,
        json={},
        headers={
            # From the original view page that is calling this sub-function
            "referer": "http://example.com/test/5/",
        },
    )
    elapsed = (monotonic_ns() - start) / 1e9
    assert response.status_code == 200
    assert response.json() == {
        "sideeffect": {
            "value_a": "Hello 1229",
        }
    }

    LOGGER.info(f"Use Experimental: {use_experimental}\nElapsed: {elapsed}")

    if min_time is not None:
        assert elapsed >= min_time
    if max_time is not None:
        assert elapsed <= max_time



from collections import defaultdict
from functools import wraps
from inspect import isclass, signature
from pathlib import Path
from typing import Any, Callable, Type

from fastapi import APIRouter, FastAPI, HTTPException, Request
from fastapi.openapi.utils import get_openapi
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from inflection import underscore
from pydantic import BaseModel
from starlette.routing import BaseRoute

from mountaineer.actions import (
    FunctionActionType,
    fuse_metadata_to_response_typehint,
    init_function_metadata,
)
from mountaineer.actions.fields import FunctionMetadata
from mountaineer.annotation_helpers import MountaineerUnsetValue
from mountaineer.config import ConfigBase
from mountaineer.controller import ControllerBase
from mountaineer.exceptions import APIException, APIExceptionInternalModelBase
from mountaineer.js_compiler.base import ClientBuilderBase
from mountaineer.js_compiler.javascript import JavascriptBundler
from mountaineer.logging import LOGGER
from mountaineer.paths import ManagedViewPath, resolve_package_path
from mountaineer.render import Metadata, RenderBase


class ControllerDefinition(BaseModel):
    controller: ControllerBase
    router: APIRouter
    # URL prefix to the root of the server
    url_prefix: str
    # Dynamically generated function that actually renders the html content
    # This is a hybrid between render() and _generate_html()
    view_route: Callable
    render_router: APIRouter

    model_config = {
        "arbitrary_types_allowed": True,
    }

    def get_url_for_metadata(self, metadata: FunctionMetadata):
        return f"{self.url_prefix}/{metadata.function_name.strip('/')}"


class ExceptionSchema(BaseModel):
    status_code: int
    schema_name: str
    schema_name_long: str
    schema_value: dict[str, Any]

    model_config = {
        "extra": "forbid",
    }


class AppController:
    """
    Main entrypoint of a project web application.

    """

    builders: list[ClientBuilderBase]
    global_metadata: Metadata | None

    def __init__(
        self,
        *,
        name: str = "Mountaineer Webapp",
        version: str = "0.1.0",
        view_root: Path | None = None,
        global_metadata: Metadata | None = None,
        custom_builders: list[ClientBuilderBase] | None = None,
        config: ConfigBase | None = None,
        fastapi_args: dict[str, Any] | None = None,
    ):
        """
        :param global_metadata: Script and meta will be applied to every
            page rendered by this application. Title will only be applied
            if the page does not already have a title set.
        :param config: Application global configuration.

        """
        self.app = FastAPI(title=name, version=version, **(fastapi_args or {}))
        self.controllers: list[ControllerDefinition] = []
        self.controller_names: set[str] = set()
        self.name = name
        self.version = version
        self.global_metadata = global_metadata
        self.builders = [
            # Default builders
            JavascriptBundler(
                environment=(
                    config.ENVIRONMENT if config is not None else "development"
                )
            ),
            # Custom builders
            *(custom_builders if custom_builders else []),
        ]

        # If this flag is present, we will re-raise this error during render()
        # so users can see the error in the browser.
        # This is useful for debugging, but should not be used in production
        self.build_exception: Exception | None = None

        # Follow our managed path conventions
        if config is not None and config.PACKAGE is not None:
            package_path = resolve_package_path(config.PACKAGE)
            self.view_root = ManagedViewPath.from_view_root(package_path / "views")
        elif view_root is not None:
            self.view_root = ManagedViewPath.from_view_root(view_root)
        else:
            raise ValueError(
                "You must provide either a config.package or a view_root to the AppController"
            )

        # The act of instantiating the config should register it with the
        # global settings registry. We keep a reference to it so we can shortcut
        # to the user-defined settings later, but this is largely optional.
        self.config = config

        self.internal_api_prefix = "/internal/api"

        # The static directory has to exist before we try to mount it
        static_dir = self.view_root.get_managed_static_dir()

        # Mount the view_root / _static directory, since we'll need
        # this for the client mounted view files
        self.app.mount(
            "/static",
            StaticFiles(directory=str(static_dir)),
            name="static",
        )

        self.app.exception_handler(APIException)(self.handle_exception)

        self.app.openapi = self.generate_openapi  # type: ignore

    def register(self, controller: ControllerBase):
        """
        Register a new controller. This will:

        - Mount the html of the controller to the main application service
        - Mount all actions (ie. @sideeffect and @passthrough decorated functions) to their public API

        :param controller: The controller instance that should be added to your webapp. The class accepts a full
        instance instead of just a class, so you're able to perform any kind of runtime initialization of the
        kwarg args that you need before it's registered.

        """
        # Since the controller name is used to build dependent files, we ensure
        # that we only register one controller of a given name
        controller_name = controller.__class__.__name__
        if controller_name in self.controller_names:
            raise ValueError(
                f"Controller with name {controller_name} already registered."
            )

        # Update the paths now that we have access to the runtime package path
        controller.resolve_paths(self.view_root, force=True)

        # The controller superclass needs to be initialized before it's
        # registered into the application
        if not hasattr(controller, "initialized"):
            raise ValueError(
                f"You must call super().__init__() on {controller} before it can be registered."
            )

        # We need to passthrough the API of the render function to the FastAPI router so it's called properly
        # with the dependency injection kwargs
        @wraps(controller.render)
        async def generate_controller_html(*args, **kwargs):
            if self.build_exception:
                raise self.build_exception

            try:
                return await controller._generate_html(
                    *args, global_metadata=self.global_metadata, **kwargs
                )
            except Exception as e:
                # If a user explicitly is raising an APIException, we don't want to log it
                if not isinstance(e, (APIExceptionInternalModelBase, HTTPException)):
                    # Forward along the exception, just modify it to include
                    # the controller name for additional context
                    LOGGER.error(
                        f"Exception encountered in {controller.__class__.__name__} rendering"
                    )
                raise

        # Strip the return annotations from the function, since we just intend to return an HTML page
        # and not a JSON response
        if not hasattr(generate_controller_html, "__wrapped__"):
            raise ValueError(
                "Unable to clear wrapped typehint, no wrapped function found."
            )

        return_model = generate_controller_html.__wrapped__.__annotations__.get(
            "return", MountaineerUnsetValue()
        )
        if isinstance(return_model, MountaineerUnsetValue):
            raise ValueError(
                "Controller render() function must have a return type annotation"
            )

        # Only the signature of the actual rendering function, not the original. We might
        # need to sniff render() again for its typehint
        generate_controller_html.__signature__ = signature(  # type: ignore
            generate_controller_html,
        ).replace(return_annotation=None)

        # Validate the return model is actually a RenderBase or explicitly marked up as None
        if not (
            return_model is None
            or (isclass(return_model) and issubclass(return_model, RenderBase))
        ):
            raise ValueError(
                "Controller render() return type annotation is not a RenderBase"
            )

        # Attach a new metadata wrapper to the original function so we can easily
        # recover it when attached to the class
        render_metadata = init_function_metadata(
            controller.render, FunctionActionType.RENDER
        )
        render_metadata.render_model = return_model

        # Register the rendering view to an isolated APIRoute, so we can keep track of its
        # the resulting router independently of the rest of the application
        # This is useful in cases where we need to do a render()->FastAPI lookup
        view_router = APIRouter()
        view_router.get(controller.url)(generate_controller_html)
        self.app.include_router(view_router)

        # Create a wrapper router for each controller to hold the side-effects
        controller_api = APIRouter()
        controller_url_prefix = (
            f"{self.internal_api_prefix}/{underscore(controller.__class__.__name__)}"
        )
        for _, fn, metadata in controller._get_client_functions():
            openapi_extra: dict[str, Any] = {
                "is_raw_response": metadata.get_is_raw_response()
            }

            if not metadata.get_is_raw_response():
                # We need to delay adding the typehint for each function until we are here, adding the view. Since
                # decorators run before the class is actually mounted, they're isolated from the larger class/controller
                # context that the action function is being defined within. Here since we have a global view
                # of the controller (render function + actions) this becomes trivial
                metadata.return_model = fuse_metadata_to_response_typehint(
                    metadata, render_metadata.get_render_model()
                )

                # Update the signature of the internal function, which fastapi will sniff for the return declaration
                # https://github.com/tiangolo/fastapi/blob/a235d93002b925b0d2d7aa650b7ab6d7bb4b24dd/fastapi/dependencies/utils.py#L207
                method_function: Callable = fn.__func__  # type: ignore
                method_function.__signature__ = signature(method_function).replace(  # type: ignore
                    return_annotation=metadata.return_model
                )

                # Pass along relevant tags in the OpenAPI meta struct
                # This will appear in the root key of the API route, at the same level of "summary" and "parameters"
                if metadata.get_media_type():
                    openapi_extra["media_type"] = metadata.get_media_type()

            controller_api.post(
                f"/{metadata.function_name}", openapi_extra=openapi_extra
            )(fn)

        # Originally we tried implementing a sub-router for the internal API that was registered in the __init__
        # But the application greedily copies all contents from the router when it's added via `include_router`, so this
        # resulted in our endpoints not being seen even after calls to `.register(). We therefore attach the new
        # controller router directly to the application, since this will trigger a new copy of the routes.
        self.app.include_router(
            controller_api,
            prefix=controller_url_prefix,
        )

        LOGGER.debug(f"Did register controller: {controller_name}")

        controller_definition = ControllerDefinition(
            controller=controller,
            router=controller_api,
            view_route=generate_controller_html,
            url_prefix=controller_url_prefix,
            render_router=view_router,
        )
        controller.definition = controller_definition

        self.controllers.append(controller_definition)
        self.controller_names.add(controller_name)

    async def handle_exception(self, request: Request, exc: APIException):
        return JSONResponse(
            status_code=exc.status_code,
            content=exc.internal_model.model_dump(),
        )

    def generate_openapi(self, routes: list[BaseRoute] | None = None):
        """
        Bundle custom user exceptions in the OpenAPI schema. By default
        endpoints just include the 422 Validation Error, but this allows
        for custom derived user methods.

        """
        openapi_base = get_openapi(
            title=self.name,
            version=self.version,
            routes=(routes if routes is not None else self.app.routes),
        )

        #
        # Exception injection
        #

        # Loop over the registered controllers and get the action exceptions
        exceptions_by_url: dict[str, list[ExceptionSchema]] = {}
        for controller_definition in self.controllers:
            for (
                _,
                _,
                metadata,
            ) in controller_definition.controller._get_client_functions():
                url = controller_definition.get_url_for_metadata(metadata)
                # Not included in the specified routes, we should ignore this controller
                if url not in openapi_base["paths"]:
                    continue

                exceptions_models = metadata.get_exception_models()
                if not exceptions_models:
                    continue

                exceptions_by_url[url] = [
                    self._format_exception_model(exception_model)
                    for exception_model in exceptions_models
                ]

        # Users are allowed to reference the same schema name multiple times so long
        # as they have the same value. If they use conflicting values we'll have
        # to use the long name instead of the short module name to avoid conflicting
        # schema definitions.
        schema_names_to_long: defaultdict[str, set[str]] = defaultdict(set)
        for exception_payloads in exceptions_by_url.values():
            for payload in exception_payloads:
                schema_names_to_long[payload.schema_name].add(payload.schema_name_long)

        duplicate_schema_names = {
            schema_name
            for schema_name, schema_name_longs in schema_names_to_long.items()
            if len(schema_name_longs) > 1
        }

        for url, exception_payloads in exceptions_by_url.items():
            existing_status_codes: set[int] = set()

            for payload in exception_payloads:
                # Validate the exception state doesn't override existing values
                # Status codes are local to this particular endpoint but schema names
                # are global because they're placed in the global components section
                if payload.status_code in existing_status_codes:
                    raise ValueError(
                        f"Duplicate status code {payload.status_code} for {url}"
                    )

                schema_name = (
                    payload.schema_name
                    if payload.schema_name not in duplicate_schema_names
                    else payload.schema_name_long
                )

                other_definitions = {
                    definition_name: self._update_ref_path(definition)
                    for definition_name, definition in payload.schema_value.pop(
                        "$defs", {}
                    ).items()
                }
                openapi_base["components"]["schemas"].update(other_definitions)
                openapi_base["components"]["schemas"][
                    schema_name
                ] = self._update_ref_path(payload.schema_value)

                # All actions are "posts" by definition
                openapi_base["paths"][url]["post"]["responses"][
                    str(payload.status_code)
                ] = {
                    "description": f"Custom Error: {payload.schema_name}",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": f"#/components/schemas/{schema_name}"}
                        }
                    },
                }

                existing_status_codes.add(payload.status_code)

        return openapi_base

    def _format_exception_model(self, model: Type[APIException]) -> ExceptionSchema:
        # By default all fields are optional. Since we are sending them
        # from the server we are guaranteed they will either be explicitly
        # provided or fallback to their defaults
        json_schema = model.InternalModel.model_json_schema()
        json_schema["required"] = list(json_schema["properties"].keys())

        return ExceptionSchema(
            status_code=model.status_code,
            schema_name=model.InternalModel.__name__,
            schema_name_long=f"{model.InternalModel.__module__}.{model.InternalModel.__name__}",
            schema_value=json_schema,
        )

    def _update_ref_path(self, schema: Any):
        """
        The $ref values that come out of the model schema are tied to #/defs instead
        of the #/components/schemas. This function updates the schema to use the
        correct prefix for the final OpenAPI schema.

        """
        if isinstance(schema, dict):
            new_schema: dict[str, Any] = {}
            for key, value in schema.items():
                if key == "$ref":
                    schema_name = value.split("/")[-1]
                    new_schema[key] = f"#/components/schemas/{schema_name}"
                    continue
                elif key == "additionalProperties":
                    # If the value is "False", we need to remove the key
                    if value is False:
                        continue

                new_schema[key] = self._update_ref_path(value)
            return new_schema
        elif isinstance(schema, list):
            return [self._update_ref_path(value) for value in schema]
        else:
            return schema

    def definition_for_controller(
        self, controller: ControllerBase
    ) -> ControllerDefinition:
        for controller_definition in self.controllers:
            if controller_definition.controller == controller:
                return controller_definition
        raise ValueError(f"Controller {controller} not found")



from dataclasses import fields, is_dataclass
from inspect import isclass
from types import UnionType
from typing import (
    Any,
    Type,
    Union,
    get_args,
    get_origin,
)

from pydantic import BaseModel
from pydantic._internal._typing_extra import eval_type_lenient


def get_value_by_alias(model: BaseModel | dict[str, Any], alias: str):
    """
    Get the value of a pydantic model by its original JSON key. This will look for both the cast
    name and the original name.

    If there's a tie, the cast name will win.

    """
    if isinstance(model, dict):
        # Dictionaries can't have aliases
        return model[alias]

    try:
        return getattr(model, alias)
    except AttributeError:
        # Only run the following if we're not able to find the cast name since in involves
        # an O(n) operation
        for field_name, field in model.model_fields.items():
            if field.alias == alias:
                return getattr(model, field_name)
    raise AttributeError(
        f"No key `{alias}` found in model, either alias or cast value."
    )


def resolve_forwardrefs(
    current_type: type,
    *,
    _globals: dict[str, Any] | None = None,
    _locals: dict[str, Any] | None = None,
):
    """
    Resolves forwardrefs to their true value. Unlike the standard eval_type_lenient logic this also
    supports nested forwardrefs like those found in origin/arg pairs.

    """
    origin = get_origin(current_type)
    args = get_args(current_type)

    if origin:
        origin = resolve_forwardrefs(origin, _globals=_globals, _locals=_locals)
        args = tuple(
            [
                resolve_forwardrefs(arg, _globals=_globals, _locals=_locals)
                for arg in args
            ]
        )

        # Workaround for UnionType not allowing programatic construction
        if origin == UnionType:
            return Union[*args]  # type: ignore

        return origin[*args]

    return eval_type_lenient(current_type, _globals or globals(), _locals or locals())


def yield_all_subtypes(
    model: type,
    _globals: dict[str, Any] | None = None,
    _locals: dict[str, Any] | None = None,
):
    """
    Given a model declaration, yield all of its subtypes. This is useful to determine whether
    a nested model might include a given subtype.

    We support:
        - Pydantic models
        - Dataclasses
        - Origin/Argument typevars, like `list[str]` or `Union[str, int]`
    """

    # Track the models we've already validated to avoid circular dependencies
    already_validated: set[Type[BaseModel]] = set()

    def resolve_types(current_type: type):
        nonlocal already_validated

        # Always echo back the current type to make sure that everything that we've processed
        # is included in the final list
        yield resolve_forwardrefs(current_type, _globals=_globals, _locals=_locals)

        if isclass(current_type) and issubclass(current_type, BaseModel):
            if current_type in already_validated:
                # Avoid circular dependencies
                # If we've already checked this class, we know it's valid
                return

            # Avoid recursion before we short-circuit any future validations to this same model
            already_validated.add(current_type)

            for field_name, field in current_type.model_fields.items():
                # Always field the full annotation, including ones with origins/args
                if field.annotation:
                    yield from resolve_types(field.annotation)

                # In the case of generics, we also want to iterate over the subvalues
                origin = get_origin(field.annotation)
                args = get_args(field.annotation)
                if origin:
                    yield from resolve_types(origin)
                    for arg in args:
                        yield from resolve_types(arg)

        elif is_dataclass(current_type):
            for dataclass_definition in fields(current_type):
                yield from resolve_types(dataclass_definition.type)

    yield from resolve_types(model)


class MountaineerUnsetValue:
    """
    A sentinel value to indicate that a value has not been set. This is useful for cases where
    we need to separate explicit Nones from missing values.
    """

    pass



from pathlib import Path
from threading import Thread
from time import sleep

import pytest

from mountaineer.watch import (
    CallbackDefinition,
    CallbackEvent,
    CallbackMetadata,
    CallbackType,
    ChangeEventHandler,
    PackageWatchdog,
)


@pytest.mark.parametrize(
    "path, ignore_list, ignore_hidden, expected_ignore",
    [
        ("myproject/regular.py", ["__pycache__"], True, False),
        ("myproject/.hidden", [""], True, True),
        ("myproject/.hidden/subfile", [""], True, True),
        ("myproject/.hidden/subfile", [""], False, False),
        ("myproject/__cache__", ["__cache__"], True, True),
    ],
)
def test_ignore_path(
    path: str,
    ignore_list: list[str],
    ignore_hidden: bool,
    expected_ignore: bool,
):
    handler = ChangeEventHandler(
        [], ignore_list=ignore_list, ignore_hidden=ignore_hidden
    )
    assert handler.should_ignore_path(Path(path)) == expected_ignore


@pytest.mark.parametrize(
    "paths, expected_paths",
    [
        (
            # Unique paths
            ["myproject/regular.py", "myproject/.hidden"],
            ["myproject/regular.py", "myproject/.hidden"],
        ),
        (
            # Simple subdirectory
            ["myproject", "myproject/subdir"],
            ["myproject"],
        ),
        (
            # More specific directory first
            ["myproject/subdir1/subdir2", "myproject/subdir1"],
            ["myproject/subdir1"],
        ),
    ],
)
def test_merge_paths(paths: list[str], expected_paths: list[str]):
    handler = PackageWatchdog("mountaineer", [])
    assert set(handler.merge_paths(paths)) == {
        str(Path(path).absolute()) for path in expected_paths
    }


def test_file_notification(tmp_path: Path):
    callback_events: list[CallbackEvent] = []

    def receive_callback(metadata: CallbackMetadata):
        callback_events.extend(metadata.events)

    def test_file_lifecycle():
        # Make sleeps longer than the debounce interval (0.1s)
        sleep(0.15)
        (tmp_path / "test.txt").write_text("Original")
        sleep(0.15)
        (tmp_path / "test.txt").write_text("Modified")
        sleep(0.15)
        (tmp_path / "test.txt").unlink()
        sleep(0.15)

        assert handler.observer
        handler.observer.stop()

    handler = PackageWatchdog(
        "mountaineer",
        [],
        callbacks=[
            CallbackDefinition(
                action=CallbackType.CREATED
                | CallbackType.MODIFIED
                | CallbackType.DELETED,
                callback=receive_callback,
            )
        ],
    )

    # Override the paths found from the package name with our temporary path
    # where we can write additional files
    handler.paths = [str(tmp_path)]

    lifecycle_thread = Thread(target=test_file_lifecycle)
    lifecycle_thread.start()

    handler.start_watching()

    event_actions: list[CallbackType] = []
    for event in callback_events:
        if event.action not in event_actions:
            event_actions.append(event.action)

    assert event_actions == [
        CallbackType.CREATED,
        CallbackType.MODIFIED,
        CallbackType.DELETED,
    ]



from . import core as CoreDependencies  # noqa: F401
from .base import (
    DependenciesBase as DependenciesBase,
    get_function_dependencies as get_function_dependencies,
    isolate_dependency_only_function as isolate_dependency_only_function,
)



from random import random, seed

import pytest

from mountaineer.cache import extended_lru_cache


@extended_lru_cache(maxsize=100)
def get_random_data(param: int):
    return random()


@pytest.fixture(autouse=True)
def reset_cache():
    get_random_data._cache.clear()  # type: ignore


def test_extended_lru_cache():
    seed(20)
    val1 = get_random_data(0)
    val2 = get_random_data(0)

    assert val1 == val2


def test_extended_lru_cache_different_params():
    seed(20)
    val1 = get_random_data(0)
    val2 = get_random_data(1)

    assert val1 != val2


def test_extended_lru_no_cache():
    seed(20)
    val1 = get_random_data(0, use_cache=False)
    val2 = get_random_data(0, use_cache=False)

    assert val1 != val2



import subprocess
from os import environ
from pathlib import Path
from shutil import which

from click import secho

from create_mountaineer_app.environments.base import EnvironmentBase


class VEnvEnvironment(EnvironmentBase):
    """
    Use the python bundled venv module to create a virtual environment
    for the current project.

    """

    def __init__(self, venv_name: str = "venv"):
        super().__init__()
        self.venv_name = venv_name

    def has_provider(self):
        return which("python3") is not None

    def install_project(self, project_path: Path):
        # Create a virtual environment
        venv_path = project_path / self.venv_name
        subprocess.run(["python3", "-m", "venv", str(venv_path)], check=True)

        secho(f"Virtual environment created at: {venv_path}", fg="green")

        # Install packages using pip
        subprocess.run(
            [str(venv_path / "bin" / "pip"), "install", "-e", "."],
            check=True,
            cwd=project_path,
            # Install within our virtualenv
            env={"PATH": f"{venv_path}/bin:{environ['PATH']}", **self.global_env},
        )

        secho("Packages installed.", fg="green")

    def run_command(self, command: list[str], path: Path):
        venv_path = path / self.venv_name
        if not venv_path.exists():
            raise ValueError(f"Virtual environment not found at: {venv_path}")

        return subprocess.Popen(
            command,
            cwd=path,
            env={"PATH": f"{venv_path}/bin:{environ['PATH']}", **self.global_env},
        )

    def get_env_path(self, project_path: Path) -> str:
        venv_path = project_path / self.venv_name
        if not venv_path.exists():
            raise ValueError(f"Virtual environment not found at: {venv_path}")

        return str(venv_path)



from re import sub as re_sub

import pytest
from fastapi import APIRouter
from fastapi.openapi.utils import get_openapi
from pydantic import BaseModel

from mountaineer.client_builder.build_actions import OpenAPIToTypescriptActionConverter
from mountaineer.client_builder.openapi import (
    ActionDefinition,
    ActionType,
    ContentBodyDefinition,
    ContentDefinition,
    OpenAPIProperty,
    OpenAPISchemaType,
    ParameterLocationType,
    URLParameterDefinition,
)
from mountaineer.constants import STREAM_EVENT_TYPE


def test_convert():
    """
    Test the conversion of a full OpenAPI schema. We make sure that the pipeline
    ends with the expected multiple fields and import, but we don't test
    the exact code generation.

    """

    class ExampleModel(BaseModel):
        pass

    class ExampleResponseModel(BaseModel):
        pass

    def fn1(payload: ExampleModel):
        pass

    def fn2():
        pass

    # Each function needs a response model because we expect that all @sideeffect
    # and @passthrough functions will have an automatically defined response model
    router = APIRouter()
    router.post("/fn1", response_model=ExampleResponseModel)(fn1)
    router.post("/fn2", response_model=ExampleResponseModel)(fn2)

    builder = OpenAPIToTypescriptActionConverter()
    openapi_spec = get_openapi(title="", version="", routes=router.routes)
    fetch_definitions, import_definitions = builder.convert(openapi_spec)

    # Returns the error classes alongside the route definitions
    assert set(fetch_definitions.keys()) == {"fn1", "fn2", "HTTPValidationError"}
    assert set(import_definitions) == {
        "ExampleModel",
        "ExampleResponseModel",
        "HTTPValidationError",
    }


EXAMPLE_REQUEST_BODY = ContentBodyDefinition(
    content_type="application/json",
    content_schema=ContentDefinition.from_meta(
        schema_ref=ContentDefinition.Reference.from_meta(
            ref="#/components/schemas/ExampleModel"
        )
    ),
)

EXAMPLE_RESPONSE_200 = ContentBodyDefinition(
    content_type="application/json",
    content_schema=ContentDefinition.from_meta(
        schema_ref=ContentDefinition.Reference.from_meta(
            ref="#/components/schemas/ExampleResponseModel"
        )
    ),
)

EXAMPLE_RESPONSE_400 = ContentBodyDefinition(
    content_type="application/json",
    content_schema=ContentDefinition.from_meta(
        schema_ref=ContentDefinition.Reference.from_meta(
            ref="#/components/schemas/HTTPValidationError"
        )
    ),
)


@pytest.mark.parametrize(
    "method_name,url,definition,expected_function,expected_imports",
    [
        # Request/response based body
        (
            "my_method_fn",
            "/testing/url",
            ActionDefinition(
                action_type=ActionType.POST,
                summary="",
                operationId="",
                requestBody=EXAMPLE_REQUEST_BODY,
                responses={
                    "200": EXAMPLE_RESPONSE_200,
                    "422": EXAMPLE_RESPONSE_400,
                },
            ),
            (
                """
                export const my_method_fn = (
                    {requestBody}: {requestBody: ExampleModel}
                ): Promise<ExampleResponseModel> => {
                    return __request({
                        'method': 'POST',
                        'url': '/testing/url',
                        'errors': {
                            422: HTTPValidationErrorException
                        },
                        'body': requestBody,
                        'mediaType': 'application/json'
                    });
                }
                """
            ),
            [
                "ExampleModel",
                "ExampleResponseModel",
            ],
        ),
        # Path and query parameters
        (
            "my_method_fn",
            "/testing/{item_id}",
            ActionDefinition(
                action_type=ActionType.POST,
                summary="",
                operationId="",
                requestBody=EXAMPLE_REQUEST_BODY,
                responses={
                    "200": EXAMPLE_RESPONSE_200,
                },
                parameters=[
                    # All path parameters are required
                    URLParameterDefinition.from_meta(
                        name="item_id",
                        schema_ref=OpenAPIProperty.from_meta(
                            title="",
                            variable_type=OpenAPISchemaType.STRING,
                        ),
                        in_location=ParameterLocationType.PATH,
                        required=True,
                    ),
                    # Required query parameter
                    URLParameterDefinition.from_meta(
                        name="query_param_required_id",
                        schema_ref=OpenAPIProperty.from_meta(
                            title="",
                            variable_type=OpenAPISchemaType.STRING,
                        ),
                        in_location=ParameterLocationType.QUERY,
                        required=True,
                    ),
                    # Cookies and headers should be skipped
                    URLParameterDefinition.from_meta(
                        name="auth_cookie",
                        schema_ref=OpenAPIProperty.from_meta(
                            title="",
                            variable_type=OpenAPISchemaType.STRING,
                        ),
                        in_location=ParameterLocationType.COOKIE,
                        required=True,
                    ),
                    URLParameterDefinition.from_meta(
                        name="auth_header",
                        schema_ref=OpenAPIProperty.from_meta(
                            title="",
                            variable_type=OpenAPISchemaType.STRING,
                        ),
                        in_location=ParameterLocationType.HEADER,
                        required=True,
                    ),
                    # Optional query parameter
                    URLParameterDefinition.from_meta(
                        name="query_param_optional_id",
                        schema_ref=OpenAPIProperty.from_meta(
                            title="",
                            variable_type=OpenAPISchemaType.STRING,
                        ),
                        in_location=ParameterLocationType.QUERY,
                        required=False,
                    ),
                ],
            ),
            (
                """
                export const my_method_fn = (
                    {
                        item_id,
                        query_param_required_id,
                        query_param_optional_id,
                        requestBody
                    }: {
                        item_id: string,
                        query_param_required_id: string,
                        query_param_optional_id?: string,
                        requestBody: ExampleModel
                    }
                ): Promise<ExampleResponseModel> => {
                    return __request({
                        'method': 'POST',
                        'url': '/testing/{item_id}',
                        'path': {
                            item_id
                        },
                        'query': {
                            query_param_required_id,
                            query_param_optional_id
                        },
                        'body': requestBody,
                        'mediaType': 'application/json'
                    });
                }
                """
            ),
            [
                "ExampleModel",
                "ExampleResponseModel",
            ],
        ),
    ],
)
def test_build_action(
    url: str,
    method_name: str,
    definition: ActionDefinition,
    expected_function: str,
    expected_imports: list[str],
):
    """
    Test the building of a single action fetch body.
    """
    builder = OpenAPIToTypescriptActionConverter()
    built_function, build_imports = builder.build_action(url, definition, method_name)

    # Exact match for function contents
    assert re_sub(r"\s+", "", built_function) == re_sub(r"\s+", "", expected_function)

    # Order doesn't matter in the imports. We assume they're all coming from the locally defined
    # /models.ts source file.
    assert set(build_imports) == set(expected_imports)


@pytest.mark.parametrize(
    "method_name,url,definition,expected_function,expected_imports",
    [
        (
            "my_method_fn",
            "/testing/url",
            ActionDefinition(
                action_type=ActionType.POST,
                summary="",
                operationId="",
                requestBody=EXAMPLE_REQUEST_BODY,
                responses={
                    "200": EXAMPLE_RESPONSE_200,
                    "422": EXAMPLE_RESPONSE_400,
                },
                media_type=STREAM_EVENT_TYPE,
            ),
            (
                """
                export const my_method_fn = (
                    {requestBody}: {requestBody: ExampleModel}
                ): Promise<AsyncGenerator<ExampleResponseModel, void, unknown>> => {
                    return __request({
                        'method': 'POST',
                        'url': '/testing/url',
                        'errors': {
                            422: HTTPValidationErrorException
                        },
                        'body': requestBody,
                        'mediaType': 'application/json',
                        'eventStreamResponse': true
                    });
                }
                """
            ),
            [
                "ExampleModel",
                "ExampleResponseModel",
            ],
        ),
    ],
)
def test_build_server_side_event_action(
    url: str,
    method_name: str,
    definition: ActionDefinition,
    expected_function: str,
    expected_imports: list[str],
):
    builder = OpenAPIToTypescriptActionConverter()
    built_function, build_imports = builder.build_action(url, definition, method_name)
    assert re_sub(r"\s+", "", built_function) == re_sub(r"\s+", "", expected_function)
    assert set(build_imports) == set(expected_imports)


@pytest.mark.parametrize(
    "method_name,url,definition,expected_function,expected_imports",
    [
        (
            "my_method_fn",
            "/testing/url",
            ActionDefinition(
                action_type=ActionType.POST,
                summary="",
                operationId="",
                requestBody=EXAMPLE_REQUEST_BODY,
                responses={
                    "200": ContentBodyDefinition(
                        content_type="application/json",
                        content_schema=ContentDefinition.from_meta(
                            schema_ref=ContentDefinition.Reference.from_meta(
                                # We won't have a response_model for raw responses
                                ref=""
                            )
                        ),
                    ),
                    "422": EXAMPLE_RESPONSE_400,
                },
                is_raw_response=True,
            ),
            (
                """
                export const my_method_fn = (
                    {requestBody}: {requestBody: ExampleModel}
                ): Promise<Response> => {
                    return __request({
                        'method': 'POST',
                        'url': '/testing/url',
                        'errors': {
                            422: HTTPValidationErrorException
                        },
                        'body': requestBody,
                        'mediaType': 'application/json',
                        'outputFormat': 'raw'
                    });
                }
                """
            ),
            [
                "ExampleModel",
            ],
        ),
    ],
)
def test_build_raw_response_action(
    url: str,
    method_name: str,
    definition: ActionDefinition,
    expected_function: str,
    expected_imports: list[str],
):
    builder = OpenAPIToTypescriptActionConverter()
    built_function, build_imports = builder.build_action(url, definition, method_name)
    assert re_sub(r"\s+", "", built_function) == re_sub(r"\s+", "", expected_function)
    assert set(build_imports) == set(expected_imports)



from functools import wraps
from inspect import (
    isasyncgen,
    isasyncgenfunction,
    isawaitable,
    isgeneratorfunction,
)
from json import dumps as json_dumps
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Coroutine,
    Literal,
    ParamSpec,
    Type,
    TypeVar,
    overload,
)

from fastapi.responses import JSONResponse, Response, StreamingResponse
from pydantic import BaseModel

from mountaineer.actions.fields import (
    FunctionActionType,
    ResponseModelType,
    extract_response_model_from_signature,
    handle_explicit_responses,
    init_function_metadata,
)
from mountaineer.constants import STREAM_EVENT_TYPE
from mountaineer.exceptions import APIException

if TYPE_CHECKING:
    from mountaineer.controller import ControllerBase

P = ParamSpec("P")
R = TypeVar("R", bound=BaseModel | AsyncIterator[BaseModel] | JSONResponse | None)

RawResponseR = TypeVar("RawResponseR", bound=Response)


@overload
def passthrough(  # type: ignore
    *,
    response_model: Type[BaseModel] | None = None,  # Deprecated
    exception_models: list[Type[APIException]] | None = None,
) -> Callable[[Callable[P, R | Coroutine[Any, Any, R]]], Callable[P, Awaitable[R]]]:
    ...


@overload
def passthrough(
    *,
    raw_response: Literal[True] = True,
) -> Callable[
    [Callable[P, RawResponseR | Coroutine[Any, Any, RawResponseR]]],
    Callable[P, Awaitable[RawResponseR]],
]:
    ...


@overload
def passthrough(
    func: Callable[P, R | Coroutine[Any, Any, R]],
) -> Callable[P, Awaitable[R]]:
    ...


def passthrough(*args, **kwargs):  # type: ignore
    """
    Only functions that are explicitly marked as actions will be accessable by the frontend. The
    @passthrough decorator indicates that this function should be called by the frontend and will
    return an explicit data payload. It will NOT update the render() state of the frontend.

    Decorate functions within your ControllerBase that you want to expose. Each of these functions should specify
    a return type. Normal passthrough endpoints can return with either a `None`, a `BaseModel` object, or a
    `JSONResponse` if you need full flexibility on return headers and content structure.

    If you do return a JSONResponse note that we will handle the merging of the response for you - so
    on the client side you will still access your endpoint contents with:

    ```typescript
    const response = await serverState.my_action({});
    console.log(response.passthrough);
    ```

    Usage:

    ```python
    from pydantic import BaseModel

    class ResponseModel(BaseModel):
        pass

    class MyController(ControllerBase):
        @passthrough
        async def my_action(self) -> ResponseModel:
            ...
    ```

    :param exception_models: List of APIException subclasses that this function is known
        to throw. These will be parsed and available to frontend clients.
    :type exception_models: list[Type[APIException]] | None

    :param raw_response: If specified, you can return a generic fastapi.Response object. There's
        no constraint this endpoint returns JSON - you can return html or a custom protocol. This
        lets you treat this API as a generic POST endpoint for you to fully control the output.
    :type raw_response: bool

    :return: The response model to use for this endpoint. If a BaseModel is not provided (you pass
        a dictionary or a SQLModel object ofr instance), we will try to convert the response object
        into the proper JSON response based on your typehint.
    :rtype: BaseModel | None | fastapi.JSONResponse

    """

    def decorator_with_args(
        response_model: Type[BaseModel] | None,
        exception_models: list[Type[APIException]] | None,
        raw_response: bool | None,
    ):
        def wrapper(func: Callable):
            passthrough_model, response_type = extract_response_model_from_signature(
                func, response_model
            )

            # Ensure our function is valid as early as possible
            # Since sync functions are blocking (unless they're run in a separate thread pool), we require long-running
            # functions to be async
            if isgeneratorfunction(func):
                raise ValueError(
                    f"Only async generators are supported: Define {func} as `async def`"
                )

            # The user has defined a generator but not a response_model
            # The frontend builder needs a response model to be able to determine the type
            if (
                isasyncgenfunction(func)
                and response_type != ResponseModelType.ITERATOR_RESPONSE
            ):
                raise ValueError(
                    f"Async generator {func} must have a response_model of type AsyncIterator[BaseModel]"
                )

            @wraps(func)
            async def inner(self: "ControllerBase", *func_args, **func_kwargs):
                response = func(self, *func_args, **func_kwargs)
                if isawaitable(response):
                    response = await response

                if raw_response:
                    return response

                if isasyncgen(response):
                    return wrap_passthrough_generator(response)

                return handle_explicit_responses(dict(passthrough=response))

            metadata = init_function_metadata(inner, FunctionActionType.PASSTHROUGH)
            metadata.passthrough_model = passthrough_model
            metadata.exception_models = exception_models
            metadata.is_raw_response = raw_response or False
            metadata.media_type = (
                STREAM_EVENT_TYPE
                if response_type == ResponseModelType.ITERATOR_RESPONSE
                else None
            )
            return inner

        return wrapper

    if args and callable(args[0]):
        # It's used as @sideeffect without arguments
        func = args[0]
        return decorator_with_args(None, None, None)(func)
    else:
        # It's used as @passthrough(xyz=2) with arguments
        return decorator_with_args(
            response_model=kwargs.get("response_model"),
            exception_models=kwargs.get("exception_models"),
            raw_response=kwargs.get("raw_response"),
        )


def wrap_passthrough_generator(generator: AsyncIterator[BaseModel]):
    """
    Simple function to convert a generator of Pydantic values to a server-event-stream
    of text payloads.

    """

    async def generate():
        async for value in generator:
            json_payload = value.model_dump(mode="json")
            data = json_dumps(dict(passthrough=json_payload))
            yield f"data: {data}\n"

    return StreamingResponse(generate(), media_type=STREAM_EVENT_TYPE)



from mountaineer.database import (
    dependencies as DatabaseDependencies,  # noqa: F401
)
from mountaineer.database.config import DatabaseConfig as DatabaseConfig
from mountaineer.database.sqlmodel import Field as Field, SQLModel as SQLModel
from mountaineer.database.validator import DatabaseValidator as DatabaseValidator



# Re-export some core dependencies from other packages that will
# be used across projects
from fastapi import Depends as Depends

from mountaineer.actions import passthrough as passthrough, sideeffect as sideeffect
from mountaineer.app import AppController as AppController
from mountaineer.config import ConfigBase as ConfigBase
from mountaineer.controller import ControllerBase as ControllerBase
from mountaineer.dependencies import CoreDependencies as CoreDependencies
from mountaineer.exceptions import APIException as APIException
from mountaineer.js_compiler.postcss import PostCSSBundler as PostCSSBundler
from mountaineer.paths import ManagedViewPath as ManagedViewPath
from mountaineer.render import (
    LinkAttribute as LinkAttribute,
    MetaAttribute as MetaAttribute,
    Metadata as Metadata,
    RenderBase as RenderBase,
    ScriptAttribute as ScriptAttribute,
    ThemeColorMeta as ThemeColorMeta,
    ViewportMeta as ViewportMeta,
)



from create_mountaineer_app.cli import get_current_version_number


def test_get_current_version_number():
    # Before we release our package and bump the version in pyproject.toml, our version
    # will be static during local development
    assert get_current_version_number() == "0.1.0"



from typing import AsyncIterator, Iterator, Optional, Type

import pytest
from fastapi.responses import JSONResponse
from pydantic import BaseModel
from pydantic.fields import FieldInfo

from mountaineer.actions.fields import (
    FunctionActionType,
    FunctionMetadata,
    ResponseModelType,
    annotation_is_metadata,
    extract_response_model_from_signature,
    fuse_metadata_to_response_typehint,
    get_function_metadata,
)
from mountaineer.actions.sideeffect import sideeffect
from mountaineer.render import Metadata, RenderBase


class ExampleRenderModel(RenderBase):
    render_value_a: str
    render_value_b: int


class ExamplePassthroughModel(BaseModel):
    passthrough_value_a: str


def basic_compare_model_fields(
    actual: dict[str, FieldInfo],
    expected: dict[str, FieldInfo],
):
    """
    FieldInfo objects can't be compared with native Python comparitors.

    """
    assert actual.keys() == expected.keys()
    for key in actual:
        assert actual[key].annotation == expected[key].annotation
        assert actual[key].default == expected[key].default
        assert actual[key].alias == expected[key].alias
        assert actual[key].is_required() == expected[key].is_required()


@pytest.mark.parametrize(
    "metadata, render_model, expected_model_name, expected_sideeffect_fields, expected_passthrough_fields",
    [
        # Standard sideeffect with every value included in the refresh
        (
            FunctionMetadata(
                action_type=FunctionActionType.SIDEEFFECT,
                function_name="example_function",
            ),
            ExampleRenderModel,
            "ExampleFunctionResponse",
            {
                # The metadata field should not be included in the sideeffect
                # response, only the defined keys
                "render_value_a": FieldInfo.from_annotation(
                    annotation=str,
                ),
                "render_value_b": FieldInfo.from_annotation(
                    annotation=int,
                ),
            },
            {},
        ),
        # Sideeffect with masked value payload, only should include the masked value
        (
            FunctionMetadata(
                action_type=FunctionActionType.SIDEEFFECT,
                function_name="example_function",
                reload_states=tuple(
                    [
                        # At runtime this will evaluate to the render_value_a metadata definition
                        ExampleRenderModel.render_value_a,  # type: ignore
                    ]
                ),
            ),
            ExampleRenderModel,
            "ExampleFunctionResponse",
            {
                "render_value_a": FieldInfo.from_annotation(
                    annotation=str,
                ),
            },
            {},
        ),
        # Passthrough function markup, should include the passthrough value
        # and no sideeffect value
        (
            FunctionMetadata(
                action_type=FunctionActionType.PASSTHROUGH,
                function_name="example_function",
                passthrough_model=ExamplePassthroughModel,
            ),
            ExampleRenderModel,
            "ExampleFunctionResponse",
            {},
            {
                "passthrough_value_a": FieldInfo.from_annotation(
                    annotation=str,
                ),
            },
        ),
    ],
)
def test_fuse_metadata_to_response_typehint(
    metadata: FunctionMetadata,
    render_model: Type[RenderBase],
    expected_model_name: str,
    expected_sideeffect_fields: dict[str, FieldInfo],
    expected_passthrough_fields: dict[str, FieldInfo],
):
    fused_model = fuse_metadata_to_response_typehint(metadata, render_model)

    if expected_sideeffect_fields:
        assert "sideeffect" in fused_model.model_fields.keys()
        assert fused_model.model_fields["sideeffect"].annotation
        basic_compare_model_fields(
            fused_model.model_fields["sideeffect"].annotation.model_fields,
            expected_sideeffect_fields,
        )

    if expected_passthrough_fields:
        assert "passthrough" in fused_model.model_fields.keys()
        assert fused_model.model_fields["passthrough"].annotation
        basic_compare_model_fields(
            fused_model.model_fields["passthrough"].annotation.model_fields,
            expected_passthrough_fields,
        )

    assert fused_model.__name__ == expected_model_name


class ParentRender(RenderBase):
    render_value_a: str


class ChildRender(ParentRender):
    render_value_b: int


def test_fuse_metadata_to_response_typehint_inherit_render():
    class ParentController(BaseModel):
        def render(self) -> ParentRender:
            return ParentRender(render_value_a="example")

        @sideeffect(reload=(ParentRender.render_value_a,))
        def sideeffect(self) -> None:
            pass

    class ChildController(ParentController):
        def render(self) -> ChildRender:
            return ChildRender(render_value_a="example", render_value_b=1)

    model_fused = fuse_metadata_to_response_typehint(
        get_function_metadata(ChildController.sideeffect),
        ChildRender,
    )
    assert set(
        model_fused.model_json_schema()["$defs"]["SideeffectResponseSideEffect"][
            "properties"
        ].keys()
    ) == {"render_value_a"}


def test_annotation_is_metadata():
    assert annotation_is_metadata(Metadata)
    assert annotation_is_metadata(Optional[Metadata])  # type: ignore
    assert annotation_is_metadata(Metadata | None)  # type: ignore
    assert not annotation_is_metadata(str)


class ExampleModel(BaseModel):
    value: str


def test_extract_response_model_from_signature():
    def none_typehint() -> None:
        pass

    def example_typehint(self) -> ExampleModel:
        return ExampleModel(value="example")

    async def example_async_typehint(self) -> ExampleModel:
        return ExampleModel(value="example")

    def example_iterator_typehint(self) -> Iterator[ExampleModel]:
        yield ExampleModel(value="example")

    async def example_async_iterator_typehint(self) -> AsyncIterator[ExampleModel]:
        yield ExampleModel(value="example")

    def no_typehint(self):
        pass

    def explicit_starlette_response(self) -> JSONResponse:
        return JSONResponse(content=dict(value="example"))

    # Regular payload return functions
    assert extract_response_model_from_signature(
        none_typehint,
    ) == (None, ResponseModelType.SINGLE_RESPONSE)
    assert extract_response_model_from_signature(
        example_typehint,
    ) == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)
    assert extract_response_model_from_signature(
        example_async_typehint,
    ) == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)

    # Iterator return functions
    assert extract_response_model_from_signature(
        example_iterator_typehint,
    ) == (ExampleModel, ResponseModelType.ITERATOR_RESPONSE)

    assert extract_response_model_from_signature(
        example_async_iterator_typehint,
    ) == (ExampleModel, ResponseModelType.ITERATOR_RESPONSE)

    assert extract_response_model_from_signature(
        explicit_starlette_response,
    ) == (None, ResponseModelType.SINGLE_RESPONSE)

    # Deprecated but test until we move support for explicit_response
    with pytest.warns(DeprecationWarning):
        assert (
            extract_response_model_from_signature(no_typehint, ExampleModel)
            == (
                ExampleModel,
                ResponseModelType.SINGLE_RESPONSE,
            )
            == (ExampleModel, ResponseModelType.SINGLE_RESPONSE)
        )

    with pytest.warns(DeprecationWarning):
        assert extract_response_model_from_signature(
            no_typehint,
        ) == (None, ResponseModelType.SINGLE_RESPONSE)



from mountaineer.controller import ControllerBase
from mountaineer.paths import ManagedViewPath
from mountaineer.render import LinkAttribute, Metadata, RenderBase
from mountaineer.views import get_core_view_path


class ExceptionRender(RenderBase):
    exception: str
    stack: str | None


class ExceptionController(ControllerBase):
    """
    Controller intended for internal use only. Allows our development server
    to render exceptions in the browser and leverage our SSR-injected live reloading.

    """

    url = "/_exception"
    view_path = (
        ManagedViewPath.from_view_root(get_core_view_path(""), package_root_link=None)
        / "core/exception/page.tsx"
    )

    def render(self, exception: str, stack: str) -> ExceptionRender:
        # Exceptions can't be passed through as API types, so parents should pre-convert them to
        # strings before rendering
        return ExceptionRender(
            exception=exception,
            stack=stack,
            metadata=Metadata(
                title=f"Exception: {exception}",
                links=[LinkAttribute(rel="stylesheet", href="/static/core_main.css")],
                ignore_global_metadata=True,
            ),
        )



from typing import Any

from mountaineer.client_builder.openapi import (
    ActionType,
    OpenAPIDefinition,
    ParameterLocationType,
)
from mountaineer.client_builder.typescript import (
    TSLiteral,
    get_typehint_for_parameter,
    python_payload_to_typescript,
)


class OpenAPIToTypescriptLinkConverter:
    """
    Take a OpenAPI definition for how a render() component needs to be accessed. Create a valid typescript
    function to format that page link with all supported parameters.

    """

    def convert(self, openapi: dict[str, Any]):
        openapi_spec = OpenAPIDefinition(**openapi)

        # We expect there to only be one path in our render router
        if len(openapi_spec.paths) != 1:
            raise ValueError(
                f"Expected only one path in render router, got {list(openapi_spec.paths.keys())}"
            )

        # Extract metadata from this GET definition to fill our link generation
        # We require a single get action
        render_payloads = [
            (render_url, endpoint)
            for render_url, render_endpoint_definition in openapi_spec.paths.items()
            for endpoint in render_endpoint_definition.actions
            if endpoint.action_type == ActionType.GET
        ]
        if len(render_payloads) != 1:
            raise ValueError(
                f"Expected exactly one GET action in render router, got {render_payloads}"
            )

        render_url, get_action = render_payloads[0]

        input_parameters: dict[str, Any] = {}
        typehint_parameters: dict[str, Any] = {}

        query_parameters: dict[str, Any] = {}
        path_parameters: dict[str, Any] = {}

        # We can now parse the query and path parameters that this endpoint supports
        for parameter in get_action.parameters:
            # We only support query and path parameters
            if parameter.in_location not in {
                ParameterLocationType.QUERY,
                ParameterLocationType.PATH,
            }:
                continue

            typehint_key, typehint_value = get_typehint_for_parameter(parameter)
            input_parameters[TSLiteral(parameter.name)] = TSLiteral(parameter.name)
            typehint_parameters[typehint_key] = typehint_value

            if parameter.in_location == ParameterLocationType.QUERY:
                query_parameters[TSLiteral(parameter.name)] = TSLiteral(parameter.name)
            elif parameter.in_location == ParameterLocationType.PATH:
                path_parameters[TSLiteral(parameter.name)] = TSLiteral(parameter.name)

        parameter_str = python_payload_to_typescript(input_parameters)
        typehint_str = python_payload_to_typescript(typehint_parameters)
        query_dict_str = python_payload_to_typescript(query_parameters)
        path_dict_str = python_payload_to_typescript(path_parameters)

        chunks: list[str] = []

        # Step 1: Parameter string with typehints
        chunks.append(
            f"export const getLink = ({parameter_str} : {typehint_str}) => {{"
        )

        # Step 2: Define our local dictionary to separate query and path parameters
        # We need to do this in the actual view controller itself because only in the backend
        # code we have an explicit delineation between query and path parameters. To the getLink
        # input function they both look the same.
        chunks.append(f"const url = `{render_url}`;")

        chunks.append(
            f"const queryParameters : Record<string, any> = {query_dict_str};\n"
            f"const pathParameters : Record<string, any> = {path_dict_str};"
        )

        # Delegate the core processing logic to our global helper since the rest of the logic
        # is shared across every link generator
        chunks.append(
            "return __getLink({rawUrl: url, queryParameters, pathParameters});"
        )

        chunks.append("};")

        return "\n\n".join(chunks)



from pathlib import Path
from re import sub as re_sub
from tempfile import TemporaryDirectory
from typing import Iterable
from unittest.mock import MagicMock, _Call, call, patch

import pytest

from mountaineer.controller import ControllerBase
from mountaineer.js_compiler.base import ClientBundleMetadata
from mountaineer.js_compiler.javascript import JavascriptBundler
from mountaineer.paths import ManagedViewPath


class MockedESBuild:
    def __init__(self):
        self.calls: list[_Call] = []

    async def bundle(self, *args, outfile: Path, **kwargs):
        # Create something that looks like a real file
        outfile.parent.mkdir(parents=True, exist_ok=True)
        outfile.write_text("FAKE ESBUILD PACKAGE CONTENTS")
        (outfile.with_suffix(".js.map").write_text("FAKE ESBUILD PACKAGE SOURCE MAP"))
        self.calls.append(call(*args, outfile=outfile, **kwargs))


@pytest.fixture
def fake_view_root() -> Iterable[ManagedViewPath]:
    with TemporaryDirectory() as temp_dir_name:
        managed_view = ManagedViewPath.from_view_root(temp_dir_name)

        # Make fake expected directories
        (managed_view / "node_modules").mkdir()
        (managed_view / "package.json").write_text("{}")
        yield managed_view


@pytest.fixture
def mocked_esbuild():
    mocked_builder = MockedESBuild()
    yield mocked_builder


@pytest.fixture(scope="function")
def base_javascript_bundler(mocked_esbuild: MagicMock) -> JavascriptBundler:
    from mountaineer.js_compiler.javascript import JavascriptBundler

    # Recycle this object every function call, since the end function will usually
    # modify the page_path or other instance variables
    return JavascriptBundler(
        root_element="root",
    )


def test_build_synthetic_endpoint(
    fake_view_root: ManagedViewPath, base_javascript_bundler: JavascriptBundler
):
    # The paths don't actually have to exist for this function
    page_path = fake_view_root / "detail" / "nested" / "page.tsx"

    (
        import_paths,
        content,
        endpoint_name,
    ) = base_javascript_bundler.build_synthetic_endpoint(
        page_path=page_path,
        layout_paths=[
            fake_view_root / "detail" / "layout.tsx",
            fake_view_root / "detail" / "nested" / "layout.tsx",
        ],
        # For ease of specifying the relative page paths, we make the test file
        # right in the root of the view directory
        output_path=fake_view_root / "dist" / "test_file.tsx",
    )

    # Should be ordered in the same hierarchy order as the layouts
    # so we can codify proper nesting behavior of layouts
    assert import_paths == [
        "import mountLiveReload from '../_server/live_reload';",
        "import Page from '../detail/nested/page';",
        "import Layout0 from '../detail/layout';",
        "import Layout1 from '../detail/nested/layout';",
    ]
    assert re_sub(r"\s+", "", content) == re_sub(
        r"\s+",
        "",
        (
            """
            const Entrypoint = () => {
                mountLiveReload({});
                return (
                    <Layout0>
                        <Layout1>
                            <Page />
                        </Layout1>
                    </Layout0>
                );
            };
            """
        ),
    )
    assert endpoint_name == "Entrypoint"


def test_build_synthetic_client_page(base_javascript_bundler: JavascriptBundler):
    # Test with simple string substitutions instead of real values so we can
    # be sure the resulting payload is dynamic.
    content = base_javascript_bundler.build_synthetic_client_page(
        synthetic_imports=[
            "SYNTHETIC_IMPORT_0",
            "SYNTHETIC_IMPORT_1",
        ],
        synthetic_endpoint="SYNTHETIC_ENDPOINT",
        synthetic_endpoint_name="SYNTHETIC_ENDPOINT_NAME",
    )

    assert re_sub(r"\s+", "", content) == re_sub(
        r"\s+",
        "",
        (
            """
            import * as React from 'react';
            import { hydrateRoot } from 'react-dom/client';
            SYNTHETIC_IMPORT_0
            SYNTHETIC_IMPORT_1

            SYNTHETIC_ENDPOINT

            const container = document.getElementById('root');
            hydrateRoot(container, <SYNTHETIC_ENDPOINT_NAME />);
            """
        ),
    )


def test_build_synthetic_ssr_page(base_javascript_bundler: JavascriptBundler):
    content = base_javascript_bundler.build_synthetic_ssr_page(
        synthetic_imports=[
            "SYNTHETIC_IMPORT_0",
            "SYNTHETIC_IMPORT_1",
        ],
        synthetic_endpoint="SYNTHETIC_ENDPOINT",
        synthetic_endpoint_name="SYNTHETIC_ENDPOINT_NAME",
    )

    assert re_sub(r"\s+", "", content) == re_sub(
        r"\s+",
        "",
        (
            """
            import * as React from 'react';
            import { renderToString } from 'react-dom/server';
            SYNTHETIC_IMPORT_0
            SYNTHETIC_IMPORT_1

            SYNTHETIC_ENDPOINT

            export const Index = () => renderToString(<SYNTHETIC_ENDPOINT_NAME />);
            """
        ),
    )


@pytest.mark.parametrize(
    "page_path, expected_layouts",
    [
        # These paths are assumed to be relative to our fake root view
        (
            Path("home/detail/page.tsx"),
            [
                Path("home/layout.tsx"),
                Path("home/detail/layout.tsx"),
            ],
        ),
        (
            Path("auth/page.tsx"),
            [Path("auth/layout.tsx")],
        ),
    ],
)
def test_sniff_for_layouts(
    page_path: Path,
    expected_layouts: list[Path],
    base_javascript_bundler: JavascriptBundler,
    fake_view_root: ManagedViewPath,
):
    # We don't have to actually add any values here, but we
    home_directory = fake_view_root / "home"
    home_detail_directory = home_directory / "detail"
    auth_directory = fake_view_root / "auth"

    home_directory.mkdir(parents=True)
    home_detail_directory.mkdir(parents=True)
    auth_directory.mkdir(parents=True)

    # Each of these directories should have a layout.tsx file in them
    (home_directory / "layout.tsx").touch()
    (home_detail_directory / "layout.tsx").touch()
    (auth_directory / "layout.tsx").touch()

    assert base_javascript_bundler.sniff_for_layouts(
        page_path=fake_view_root / page_path,
        view_root_path=fake_view_root,
    ) == [
        (fake_view_root / relative_path).resolve().absolute()
        for relative_path in expected_layouts
    ]


@pytest.mark.asyncio
async def test_convert(
    base_javascript_bundler: JavascriptBundler,
    fake_view_root: ManagedViewPath,
    mocked_esbuild: MockedESBuild,
):
    class ExampleController(ControllerBase):
        async def render(self) -> None:
            pass

    with (
        patch.object(base_javascript_bundler, "sniff_for_layouts") as sniff_for_layouts,
        patch.object(
            base_javascript_bundler, "build_synthetic_endpoint"
        ) as build_synthetic_endpoint,
        patch.object(
            base_javascript_bundler, "build_synthetic_client_page"
        ) as build_synthetic_client_page,
        patch.object(
            base_javascript_bundler, "build_synthetic_ssr_page"
        ) as build_synthetic_ssr_page,
    ):
        sniff_for_layouts.return_value = [
            fake_view_root / "layout.tsx",
            fake_view_root / "nested/layout.tsx",
        ]

        build_synthetic_endpoint.return_value = (
            ["import Page from './page';", "import Layout from './layout';"],
            "const Entrypoint = () => { return <Layout><Page /></Layout>; };",
            "Entrypoint",
        )

        build_synthetic_client_page.return_value = "CLIENT_PAGE"
        build_synthetic_ssr_page.return_value = "SSR_PAGE"

        await base_javascript_bundler.start_build()

        await base_javascript_bundler.handle_file(
            file_path=fake_view_root / "page.tsx",
            controller=ExampleController(),
            metadata=ClientBundleMetadata(),
        )

        await base_javascript_bundler.finish_build()

        ssr_path = (
            fake_view_root.get_managed_ssr_dir(tmp_build=True) / "example_controller.js"
        )
        static_paths = {
            (path, ".map" in path.name)
            for path in fake_view_root.get_managed_static_dir(tmp_build=True).iterdir()
            if path.is_file() and "example_controller" in path.name
        }

        assert ssr_path.exists()
        assert ssr_path.with_suffix(".js.map").exists()

        assert "SSR_PAGE" in ssr_path.read_text()

        assert len(static_paths) == 2
        static_path = next(path for path, is_map in static_paths if not is_map)

        assert "CLIENT_PAGE" in static_path.read_text()


@pytest.mark.parametrize(
    "page_path, view_root_path, expected_throws",
    [
        # Valid page path and relative location
        (
            Path("home/detail/page.tsx"),
            Path("home"),
            False,
        ),
        # Similar path but different at the end
        (
            Path("home/detail/page.tsx"),
            Path("home/detail/nested"),
            True,
        ),
        # Totally separate path
        (
            Path("home/detail/page.tsx"),
            Path("other_path"),
            True,
        ),
        # Nested but not a page
        (
            Path("home/detail/nested/page.txt"),
            Path("home/detail/nested"),
            True,
        ),
    ],
)
def test_validate_nested_paths(
    page_path: Path,
    view_root_path: Path,
    expected_throws: bool,
    base_javascript_bundler: JavascriptBundler,
):
    if expected_throws:
        with pytest.raises(ValueError):
            base_javascript_bundler.validate_page(
                page_path=page_path, view_root_path=view_root_path
            )
    else:
        base_javascript_bundler.validate_page(
            page_path=page_path, view_root_path=view_root_path
        )
